import sys
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtGui import QPainter,QPainterPath
from PyQt5 import QtGui
from PyQt5.QtCore import Qt,QSizeF
from .my_interpolation import my_interpolate
# from scipy.interpolate import interp1d
import time
from PyQt5.QtWidgets import QFileDialog,QMessageBox
from PyQt5.QtGui import QPixmap
from PyQt5 import QtPrintSupport
from PyQt5.QtPrintSupport import QPrintDialog, QPrinter
from PyQt5.QtGui import QFont, QFontDatabase

import os
import numpy as np
import traceback 

# from transaction_editor import TransactionEditor
# from agent_editor import AgentEditor

from .agent_editor import CodeEditor


class Label:
    """
    A label for a box
    """

    label_instances = {}


    def __new__(cls,x,y,name,subject,parent_box):
        print("NEW", cls,x,y,name,subject)

        if name in cls.label_instances:
            if cls.label_instances[name] is not None:

                return cls.label_instances[name]

        return super().__new__(Label)



    def __init__(self,x,y,name,subject,parent_box):
        self.x = x
        self.y =  y
        self.box = parent_box
        self._name = name
        self.subject = subject

        self.offx = 0
        self.offy = 0

        self.__class__.label_instances[name] = self

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self,val):
        self.__class__.label_instances[val] = self
        self.__class__.label_instances[self._name] = None
        self._name = val


    def draw(self, qp, brush, position = None):
        #
        #
    	
        dw = MyDrawWidget.instance
        zf = dw.zoom_factor

        # set the name according to the checkbox settings
        if not dw.main_widget.checkBox_8.isChecked():
            name = self._name.replace("_"," ")
        else:
            name = self.subject.replace("_"," ")

        # check the actually dispalyed height and length
        fm = qp.fontMetrics()

        text_length = fm.boundingRect(name).width() + 2
        text_height = fm.boundingRect(name).height() + 2
    	
        if position is None:
            a = zf*(dw.gox + self.x + self.offx )
            b = zf*(dw.goy + self.y + self.offy )
        else:
            a,b = position

        pw = MyDrawWidget.instance # self.box.parent_widget
        # draw highlighted background
        if brush is not None and dw.main_widget.checkBox_4.isChecked():
            px = (int(a-.1))
            py = (int(b-text_height+1))

            qx = (int(a + text_length))
            qy = (int(b))

            bound_rect = QtCore.QRectF(QtCore.QPointF(px,py),QtCore.QPointF(qx,qy))
            qp.fillRect(bound_rect,brush)

        # draw text
        # qp.drawText(pw.raster(int(a)),pw.raster(int(b)-1), name)
        qp.drawText(int(a), int(b-1), name)


class Box:
    """
    A box (representing an agent)
    """
    def __init__(self, x, y, name, subject, ishelper=False, parent=None, parent_widget=None, is_support=False):

        self._x = x
        self._y = y

        self.x0 = x
        self.y0 = y

        self._name = name #.capitalize()
        self.subject = subject # if is a helper box, subject is the transaction subject
        self.label = Label(x+1,y-7,name,subject,parent_box = self)

        self.ishelper = ishelper
        self.is_support = is_support
        self.parent = parent
        self.n_connections = 0

        self.parent_widget = parent_widget

    @property
    def name(self):
        return self._name
    @name.setter
    def name(self,val):
        self._name = val
        self.label.name = val

    @property
    def x(self):
        if self.parent_widget is not None:
            return self.parent_widget.raster(self._x)
        
        return self._x

    @property
    def y(self):
        if self.parent_widget is not None:
            return self.parent_widget.raster(self._y)

        return self._y

    @x.setter
    def x(self,val):
        dx = val - self._x
        self._x += dx
        self.label.x += dx

    @y.setter
    def y(self,val):
        dy = val - self._y
        self._y += dy
        self.label.y += dy

    def draw(self, qp, brush=None, active=True):

        dw = MyDrawWidget.instance
        zf = dw.zoom_factor

        if self.ishelper:

            if active:
                handle_w = 3
            else:
                handle_w = 1

            pw = MyDrawWidget.instance # self.parent_widget
            px = int(zf*(dw.gox + self.x - .5*handle_w))# - handle_w)
            py = int(zf*(dw.goy + self.y - .5*handle_w))# - handle_w)

            #if pw is not None: 
            #    px = pw.raster(px) 
            #    py = pw.raster(py) 
            
            qp.drawEllipse(QtCore.QPointF(px,py), 2* handle_w, 2* handle_w)

        else:

            pw = self.parent_widget
            path = QPainterPath()

            w = 45
            x = self.x
            y = self.y
            x2 = self.x + w
            y2 = self.y+ w

            x1 = zf*(dw.gox +x)
            y1 = zf*(dw.goy +y)
            x2 = zf*(dw.gox +x2)
            y2 = zf*(dw.goy +y2)

            # rect with rounded corners
            round_w = 4

            # draw boxes
            if pw is not None:
                x1 = pw.raster(x1)
                y1 = pw.raster(y1) 

            path.addRoundedRect(QtCore.QRectF(x1,y1,zf*w,zf*w), zf*round_w, zf*round_w)
            qp.fillPath(path,brush.color())


    def edit_agent(self):
        if not self.ishelper:
            print("EDIT AGENT CODE", self.name)
            print("INSTANCES",CodeEditor.instances,"\n\n")
            start = ""
            if self.name in self.parent_widget.code_data:
                start = self.parent_widget.code_data[self.name]

            # workaround for old file versions
            if self.name.lower() in self.parent_widget.code_data:
                start = self.parent_widget.code_data[self.name.lower()]

            if self.name not in CodeEditor.instances:
                dw = self.parent_widget.main_widget.dockWidget_7
                tw = self.parent_widget.main_widget.tabWidget

                pos = self.parent_widget.main_widget.pos()
                x = pos.x()
                y = pos.y()
                w = self.parent_widget.main_widget.frameGeometry().width()
                h = self.parent_widget.main_widget.frameGeometry().width()

                tab = QtWidgets.QWidget()

                scrollbar = QtWidgets.QScrollArea(widgetResizable=True)
                scrollbar.setWidget(tab)
                #tab.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,QtWidgets.QSizePolicy.Expanding))

                print("New code editor in ",dw,"parent",dw.parent())
                if self.parent_widget.main_widget.current_file is None:
                    self.parent_widget.main_widget.notify("Please save the project before editing agents.",title="Save Project")
                else:
                    ce = CodeEditor(tab,self,start)
                    tw.addTab(tab, ce.search_name())
                    ce.tab = tab

                    tab_idx = tw.indexOf(tab)

                    #ce.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,QtWidgets.QSizePolicy.Expanding))
                    print("New code editor in ",dw,"parent",dw.parent())
                    ce.setStyleSheet(dw.parent().theme_manager.get_stylesheet("main"))
                    ce.setStyleSheet(dw.parent().theme_manager.get_background_style())
                    ce.setGeometry(2,2,self.parent_widget.frameGeometry().width()-2,self.parent_widget.frameGeometry().height()-2)
                    tw.setGeometry(0,0,self.parent_widget.frameGeometry().width(),self.parent_widget.frameGeometry().height())
                    dw.dockLocationChanged.connect(ce.resizeEvent)

                    tw.setCurrentIndex(tab_idx)
                    ce.resize_frame()

            # activate current tab
            if self.name in CodeEditor.instances:
                current_tab = CodeEditor.instances[self.name].tab
                tw = self.parent_widget.main_widget.tabWidget
                tw.setCurrentWidget(current_tab)
                tw.setTabText(tw.indexOf(current_tab),self.name)

        else:
            print("CANNOT EDIT HELPER")

    def overlaps(self,boxes):
        if self.ishelper:
            e = 8
        else:
            return False

        for box in boxes:

            if self.x - box.x < e:
                return True
            if self.y - box.y < e:
                return True

        return False

    def adjust_position(self, boxes):
        k = 0
        while self.overlaps(boxes) and k < 100:

            self.x = self.x0 + np.random.uniform(-45,45)
            self.y = self.y0 + np.random.uniform(-45,45)

            try:
                self.x = np.clip(self.x ,1,self.parent_widget.W - 45) # 799)
                self.y = np.clip(self.y ,1,self.parent_widget.H -45) # 699)
                # TODO fix parent_widget is None sometimes
            except:
                pass

            k+= 1


class Connector:
    def __init__(self,a,b,name="Connector",items=None,subject="(None)"):
        """
        creates a new connector

        :param a: a box (a) sender
        :param b: a box (b) receiver
        :param name: the label of the conncetor
        :param items: items affected by this connector (list of str)
        """

        self.a = a
        self.b = b

        if items is None:
            self.items = []
        else:
            self.items =items

        self.subject = subject

        #self.a.n_connections += 1
        #if self.a != self.b:
        #    self.b.n_connections += 1
        # ^moved to add_connection

        dx = self.b.x +25/2. - self.a.x #+ 25
        dy = self.b.y +25/2. - self.a.y #+ 25

        u = np.random.uniform(-1,1)
        v = np.random.uniform(-1,1)

        if a != b:
            #self.c = Box(self.a.x + 0.5*dx + u, self.a.y + 0.5*dy + v,name,subject,ishelper=True,parent=self)
            self.c = Box(self.a.x + 0.5*dx + u, self.a.y + 0.5*dy + v,name,subject,ishelper=True,parent=self)
            self.d = Box(self.a.x + 0.75*dx, self.a.y + 0.75*dy,name+"<support>",subject,ishelper=True,parent=self,is_support=True)

            #self.d = Box(self.a.x + 0.75*dx, self.a.y + 0.75*dy,name+"<support>",subject,ishelper=True,parent=self,is_support=True) # <- is support handle for better line interpolation
        else:
            self.c = Box(self.a.x + 5*u, self.a.y + 5*u ,name,subject,ishelper=True,parent=self)
            self.d = Box(self.a.x + 0.75*dx, self.a.y + 0.75*dy,name+"<support>",subject,ishelper=True,parent=self,is_support=True)

        self.a.ishelper = False
        self.b.ishelper = False

        # NOTE small connector boxes c,d are called 'helper' boxes. The
        # support the line drawn between the 'agent' boxes a, b


class MyDrawWidget(QtWidgets.QWidget):

    instance = None

    def __init__(self,parent_widget,main_widget):

        super().__init__(parent_widget)

        parent_widget.resizeEvent = self.resizeEvent
        # self.W = int(self.parent().frameGeometry().width()-1030) # 850
        # self.H = 750
        # self.setGeometry(0,0,self.W,self.H)
        self.move(0,0)

        self.parent_widget = parent_widget
        self.main_widget = main_widget

        self.resize_frame()
        self.__class__.instance = self

        self.boxes = []
        self.connectors = []

        self.old_positions = {}

        self.code_data = {}
        self.box_position_data = {}
        self.label_position_data = {}

        self.selected = None
        self.connected1 = None
        self.connected2 = None
        self._highlighted = None

        self.last_click = time.time()

        self.main_widget.zoomOutButton.pressed.connect(self.zoom_out)
        self.main_widget.zoomInButton.pressed.connect(self.zoom_in)

        self.main_widget.graphEditButton.setEnabled(False)

        self.setMouseTracking(True)

        self.W0 =  max(400,int(self.parent_widget.frameGeometry().width()-190)) # 850
        self.H0 =  max(1000,int(self.parent_widget.frameGeometry().height()-100)) # 750

        self.offx = 0
        self.offy = 0 # temporary offset when moving a box1

        self.gox = 0   # global offset x-axis
        self.goy = 0   # global offset y-axis
        self.go_startx = 0 # starting position when moving global offset
        self.go_starty = 0
        self.gox_start = 0
        self.goy_start = 0 # starting position before moving global offset

        self.maxx = self.W-50 # 800
        self.maxy = self.H-50 # 700

        self.mousex = 0
        self.mousey = 0

        self.interpol_method = "quadratic"
        self.mode = "select"

        self.zoom_factor = 1.0

        self.raster_size = 20.0 # 0.9
        self.show_raster = False

        self.show()


    def update_raster(self):
        # updates the raster settings

        # self.raster_size = 10.0 # self.main_widget.spinBoxRaster.value()  #
        # * 0.01
        self.show_raster = self.main_widget.checkBox_raster.isChecked()
        print("raster", self.raster_size,self.show_raster)

        self.update()

    def resizeEvent(self, event):
        # print("RESIZE DRAW CANVAS")
        self.resize_frame()

    def swap_interpol_style(self):
        if self.interpol_method == "cubic":
            self.interpol_method = "quadratic"
        elif self.interpol_method == "quadratic":
            self.interpol_method = "slinear"
        elif self.interpol_method == "slinear":
            self.interpol_method = "cubic"
        self.main_widget.update()

    def wheelEvent(self,event):

        numDegrees = event.angleDelta().y()
        numSteps = numDegrees / 15
        if numSteps > 0:
            self.zoom_in()
        else:
            self.zoom_out()
        event.accept()

    def zoom_in(self):
        self.zoom_factor += 0.1

        if self.zoom_factor > 3.0:
            self.zoom_factor = 3.0


        self.update()

    def zoom_out(self):

        self.zoom_factor -= 0.1
        if self.zoom_factor < .4:
            self.zoom_factor = .4

        self.update()


    def resize_frame(self):
        # NOTE not used from version spotykach TODO think of re-introducing this when parent widget is resized

        #self.W = max(500,int(self.parent().frameGeometry().width()-970)) # 850
        #self.H = max(850,int(850+self.parent().frameGeometry().height()-986)) # 750
        self.W = int(self.parent_widget.frameGeometry().width()-10) # 850
        self.H = int(self.parent_widget.frameGeometry().height()-10) # 750

        self.maxx = self.W- 50 # 800
        self.maxy = self.H- 50 # 700
        # self.setGeometry(890,170,self.W,self.H)
        self.setGeometry(5,13,self.W,self.H)

    def export_pdf(self):
        success = False

        if self.main_widget.current_file is not None:
            try:
                filename = QFileDialog.getSaveFileName(self, 'Open file',os.path.dirname(os.path.abspath(self.main_widget.current_file)), "PDF Files (*.pdf)")[0]
                success = True
            except:
                pass
        if not success or self.main_widget.current_file is None:
            try:
                filename = QFileDialog.getSaveFileName(self, 'Open file',os.getcwd(), "PDF Files (*.pdf)")[0]
                success = True
            except Exception as e:
                self.main_widget.notify("Something went wrong while exporting the Graph as PDF:\n" + str(e), title="Error")

        if not success or filename == "" or filename is None:
            return

        # export current's frame painting to pdf
        printer = QPrinter()
        printer.setOutputFormat(QtPrintSupport.QPrinter.PdfFormat)
        printer.setOutputFileName(filename)
        printer.setOrientation(QtPrintSupport.QPrinter.Landscape)
        printer.setFullPage(False) # keep?
        marginw = 100
        marginh = 100
        printer.setPaperSize(QSizeF(self.width()-marginw,self.height()-marginh),QPrinter.Point)
        qp = self.paint(reference=printer,ratio=1.1)
        qp.end()


    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self,val):
        if val is None:
            self.main_widget.graphEditButton.setEnabled(False)
            self.main_widget.graphDeleteButton.setEnabled(False)

            self._highlighted = val
        else:
            if not val.ishelper: # is not a helper node (connection node)
                self.main_widget.graphEditButton.setEnabled(True)
                # self.parent().graphDeleteButton.setEnabled(True)
                if val.n_connections == 0:
                    self.main_widget.graphDeleteButton.setEnabled(True)
                else:
                    print("Delete disabled because has still %i connections" % val.n_connections)
                    self.main_widget.graphDeleteButton.setEnabled(False)
            else:
                print("Cannot delete (is helper)")

                if str(val.label.name) != "None":
                    self.main_widget.graphEditButton.setEnabled(False)
                    self.main_widget.graphDeleteButton.setEnabled(False)

            self._highlighted = val

    def box_positions(self):
        pos_data = {}

        for box in self.boxes:
            pos_data[box.name] = {"x": int(box.x), "y": int(box.y)}

        return pos_data

    def label_positions(self):
        pos_data = {}

        for box in self.boxes:
            box.label.x += box.label.offx
            box.label.y += box.label.offy

            box.label.offx = 0
            box.label.offy = 0

            pos_data[box.name + "<label>"] = {"x": int(box.label.x), "y": int(box.label.y)}

        return pos_data


    def reposition(self):
        pos_data = self.old_positions # box_position_data
        # reposition all the boxes
        print("reposition <-> ", pos_data)
        for box in self.boxes:
            if box.name in pos_data:

                print("reposition",box.name,pos_data[box.name])
                box.x = pos_data[box.name][0] # ["x"]
                box.y = pos_data[box.name][1] # ["y"]
                box.x0 = box.x
                box.y0 = box.y

        for connector in self.connectors:
            for box in [connector.a, connector.b, connector.c, connector.d]:
                if box.name in pos_data:

                    print("reposition",box.name,pos_data[box.name])
                    box.x = pos_data[box.name][0] # ["x"]
                    box.y = pos_data[box.name][1] # ["y"]
                    box.x0 = box.x
                    box.y0 = box.y
                

    def reposition_labels(self):
        pos_data = self.label_position_data # label_position_data

        # reposition all the labels
        print("reposition <-> ", pos_data)

        for box in self.boxes:
            print("box", box)
            label_name = box.name + "<label>"

            #try:
            if label_name in pos_data:
                
                print("reposition", label_name, pos_data[label_name])
                box.label.x = pos_data[label_name]["x"] # ["x"]
                box.label.y = pos_data[label_name]["y"] # ["y"]
                
                box.label.offx = 0
                box.label.offy = 0
                
                # except Exception as e:
                #    self.parent_widget.notify(str(e), title="Error")

        for label in Label.label_instances.values():

            print("label", label.name)
            if label is not None:
                label_name = label.name + "<label>"

                if label_name in pos_data:
                    label.x = pos_data[label_name]["x"] # ["x"]
                    label.y = pos_data[label_name]["y"] # ["y"]

                    print("LABEL POSITIONED at" , label.x, label.y)
                    #label.offx = 0
                    #label.offy = 0
        

        print("ok.")


    def edit_agent(self):
        if self.highlighted is not None:
            self.highlighted.edit_agent()

    def remove_agent(self):
        # remove currently selected agent if it has no connections
        box = self.highlighted
        if box is not None and box.n_connections <= 0:
            if box in self.boxes:
                self.boxes.remove(box)
            if box.name in self.code_data:
                del self.code_data[box.name]

            if box.name in CodeEditor.instances:
                ce = CodeEditor.instances[box.name]
                tw = box.parent_widget.main_widget.tabWidget
                tw.removeTab(tw.indexOf(ce.tab))

            self.update()
        else:
            self.main_widget.notify("Box has still %i connection(s)! Cannot remove"%box.n_connections, title="Error")

    def check_exist(self,name):
        if name == "":
            return False

        for box in self.boxes:
            # TODO more efficient search for large number of boxes?
            if box.name == name:
                return True
        return False

    def add_agent(self, name):

        # mechanism to avoid overwriting...
        for box in self.boxes:
            if not box.ishelper and box.name == name:
                return box

        # self.parent().notify("Add agent", title="")
        
        print("add agent", name)
        print("old positions", self.old_positions)

        # actually a new agent?
        if name not in self.old_positions:
            u = np.random.uniform(40-self.gox, self.W-40-self.goy)
            v = np.random.uniform(40-self.gox ,self.H-40-self.goy)
            
            self.old_positions[name] = u,v
        else:
            u, v = self.old_positions[name]
            print("Move to old position", name, u, v)

        new_box = Box(u, v, name, subject=name, ishelper=False, parent_widget=self)

        self.boxes.append(new_box)
        
        # check label position 
        if name + "<label>" in self.old_positions:
            new_box.label.x = self.old_positions[name+"<label>"][0]
            new_box.label.y = self.old_positions[name+"<label>"][1]

        self.update()

        return new_box

    def arrange_pretty(self):
        for box in self.boxes:
            box.adjust_position(self.boxes)

        self.update()

    def highlight_connector(self,name):
        for box in self.boxes:
            if box.ishelper and box.name == name or box.name.lower() == name.lower():
                self.highlighted = box

        self.update()


    def add_connection(self, box1, box2, name, items, subject="(None)"):

        #for box in self.boxes:
        #    if box.ishelper and box.name == name:
        #        return

        new_conn = Connector(box1, box2, name, items, subject=subject)
        print("NEW CONN", name)

        if name not in self.old_positions:
            self.old_positions[name] = new_conn.c.x, new_conn.c.y
        else:
            new_conn.c.x = self.old_positions[name][0]
            new_conn.c.y = self.old_positions[name][1]

            new_conn.c.label.x = self.old_positions[name+"<label>"][0]
            new_conn.c.label.y = self.old_positions[name+"<label>"][1]
            
            print("CONN C", self.old_positions[name+"<label>"])

        if name+"<support>" not in self.old_positions:
            self.old_positions[name+"<support>"] = new_conn.d.x, new_conn.d.y
        else:
            new_conn.d.x = self.old_positions[name+"<support>"][0]
            new_conn.d.y = self.old_positions[name+"<support>"][1]
            
            new_conn.d.label.x = self.old_positions[name+"<support><label>"][0]
            new_conn.d.label.y = self.old_positions[name+"<support><label>"][1]
            
            print("CONN D", self.old_positions[name+"<support><label>"])

        box1.n_connections += 1
        box2.n_connections += 1

        self.connectors.append(new_conn)
        self.boxes.append(new_conn.c)
        self.boxes.append(new_conn.d)

        return new_conn

    def clear(self,clearall=False):

        if clearall:
            boxes = [b for b in self.boxes]

        else:
            boxes = [b for b in self.boxes if b.ishelper]

        while len(boxes) > 0:
            b = boxes.pop()
            self.boxes.remove(b)
            self.old_positions[b.name] = (b.x, b.y)
            try:
                self.old_positions[b.name + "<label>"] = (b.x, b.y)
            except:
                pass

        while len(self.connectors) > 0:
            self.connectors.pop()


    def rename_connection(self,name,new_name):
        for box in self.boxes:
            if box.ishelper and box.name == name:
                box.name = new_name
                box.label.name = new_name
                self.update()
                self.old_positions[box.name] = (box.x,box.y)

            if box.ishelper and box.name == name+"<support>":
                box.name = new_name+"<support>"
                box.label.name = new_name + "<support>"
                self.update()
                self.old_positions[box.name] = (box.x, box.y)

    def remove_connection(self,name):
        self.update()

        self.highlighted = None
        self.selected = None
        print("remove connection",name)

        for box in self.boxes:
            if box.ishelper and box.name == name:
                print("*BOX a", box.parent.a.n_connections)
                print("*BOX b", box.parent.b.n_connections)

                box.parent.a.n_connections -= 1
                box.parent.b.n_connections -= 1

                print("remove box", box, box.name)
                print("BOX a", box.parent.a.n_connections)
                print("BOX b", box.parent.b.n_connections)

                self.boxes.remove(box)
                self.connectors.remove(box.parent)

                if False:
                    # TODO ask dialog if boxes are to be deleted if they have no longer any connection to other boxes...
                    if box.parent.a.n_connections == 0:
                        self.boxes.remove(box.parent.a)
                    if box.parent.b.n_connections == 0:
                        self.boxes.remove(box.parent.b)

                # print("self.boxes",[b.name for b in self.boxes])

                self.old_positions[box.name] = (box.x, box.y)
                self.old_positions[box.parent.a.name] = (box.parent.a.x,box.parent.a.y)
                self.old_positions[box.parent.b.name] = (box.parent.b.x,box.parent.b.y)

                self.update()
                return

    def manhattan_lines(self,x0, y0, xf, yf, n=50):
    
        wx = abs(xf - x0) # abs(ip[-1][0] - ip[0][0])
        wy = abs(yf - y0) #  abs(ip[-1][1] - ip[0][1])

        eps = 1e-2
        if wx <= eps:
            x_coords = [x0  for i in range(n)]
            y_coords = list(np.linspace(y0,yf,n))
            return x_coords, y_coords
        
        if wy <= eps:
            y_coords = [y0 for i in range(n)]
            x_coords = list(np.linspace(x0,xf,n))
            return x_coords, y_coords
        
    
        len1 = int(0.75*n)
        len2 = int(0.25*n)

        if wx > wy:

            x_coords = list(np.linspace(x0,xf,len1)) + [xf]*len2
            y_coords = [y0]*len1 + list(np.linspace(y0,yf,len2))

        else:

            y_coords = list(np.linspace(y0,yf,len1)) + [yf]*len2
            x_coords = [x0]*len1 + list(np.linspace(x0,xf,len2))

        n = len(x_coords)

        ##x_coords = list(map(self.raster,x_coords))
        #y_coords = list(map(self.raster,y_coords))

        return x_coords, y_coords

    def crop_angle(self,x):
        x = x % np.pi

        if 0 <= x < np.pi/2:
            x = 0.0

        elif np.pi/2 <= x < np.pi:
            x = np.pi/2

        elif -np.pi/2 <= x < 0 :
            x = -np.pi/2.0

        elif -np.pi <= x < -np.pi/2:
            x = -np.pi

        return x

    def draw_arrow_at(self, qp, interpolated_points, md,offset_text=15, reverse = False, dashed=False,dd=5):

        if self.show_raster:
            dd = 5
            md = int(len(interpolated_points)/2.)
            # reverse = not reverse
        
        mid_point = interpolated_points[md]
        mid_point2 = interpolated_points[md+dd]
        while dd < len(interpolated_points) and (mid_point2[0] == mid_point[0] and mid_point2[1] == mid_point[1]):
            dd+= 1
            mid_point2 = interpolated_points[md+dd]
        
        mid_points3 = interpolated_points[md-offset_text+1]
        mid_points4 = interpolated_points[md-offset_text+1+2*dd]

        # for debugging:
        #qp.drawRect(int(mid_point[0]),int(mid_point[1]),8,8)
        #qp.drawRect(int(mid_point2[0]),int(mid_point2[1]),8,8)
        
        # start = interpolated_points[0] 
        # end = interpolated_points[-1]

        dy = mid_point[1] - mid_point2[1]
        dx = mid_point[0] - mid_point2[0]

        if reverse:
            dx = -dx
            dy = -dy

        alfa = np.arctan2(dy,dx)

        eps = 1e-10

        if abs(dy) < eps and abs(dx) >= 0:
            alfa = 0.0

        elif abs(dy) < eps and abs(dx) < 0:
            alfa = np.pi

        elif abs(dx) < eps and abs(dy) >= 0:
            alfa = -np.pi/2.

        elif abs(dx) < eps and abs(dy) < 0:
            alfa = np.pi/2.

        if self.show_raster:
            alfa = self.crop_angle(alfa)
        
        beta1 = alfa + np.pi/4
        beta2 = alfa - np.pi/4

        if abs(dy) > abs(dy):
            beta3 = alfa - np.pi/3
        else:
            beta3 = alfa - np.pi/3 + np.pi/2

        x2 = mid_point2[0]
        y2 = mid_point2[1]

        d = 10 # arrow width
        x3 = np.cos(beta1)*d + x2
        y3 = np.sin(beta1)*d + y2

        x4 = np.cos(beta2)*d + x2
        y4 = np.sin(beta2)*d + y2

        xt = np.cos(beta3)*3 + mid_points3[0]
        yt = np.sin(beta3)*3 + mid_points3[1]

        if self.show_raster:

            darr = 10 # arrow width
            eps = 1e-10
            if abs(dy) < eps and (dx) <= 0:
                x3 = x2 - d
                y3 = y2 + darr

                x4 = x2 - d
                y4 = y2 - darr

            elif abs(dy) < eps and (dx) > 0:
                x3 = x2 + d
                y3 = y2 + darr

                x4 = x2 + d
                y4 = y2 - darr


            elif abs(dx) < eps and (dy) >= 0:
                x3 = x2 - darr
                y3 = y2 + d

                x4 = x2 + darr
                y4 = y2 + d


            elif abs(dx) < eps and (dy) < 0:
                x3 = x2  - darr
                y3 = y2 - d

                x4 = x2 + darr
                y4 = y2 - d


        # draw arrow
        # if is_flow: # TODO think about behavior for non-flows
        if not dashed:
            if self.main_widget.checkBox.isChecked(): # not end_is_start and  ...

                # left side
                qp.drawLine(int(self.zoom_factor*(x2)),int(self.zoom_factor*(y2)), int(self.zoom_factor*(x3)), int(self.zoom_factor*(y3)))

                # right side
                qp.drawLine(int(self.zoom_factor*(x2)), int(self.zoom_factor*(y2)), int(self.zoom_factor*(x4)), int(self.zoom_factor*(y4)))

        return xt, yt


    def curved_connector(self,qp,x0,y0,xf,yf,xi,yi,xd,yd,dashed=False,is_flow=True,is_unidir=True):
        """
        draws a curved connector
        """

        #if dashed:
        #   return 0,0 # TODO maybe add another checkbox to configure behavior for non-active connectors

        # x0 += 12.5
        # y0 += 12.5

        # construct helper box_positions
        xi2 = 0.5*(xi-x0) + x0
        xi3 = 0.5*(xf-xi) + xi
        yi2 = 0.5*(yi-y0) + y0
        yi3 = 0.5*(yf-yi) + yi

        #qp.drawRect(xi2-5,yi2-5,10,10)
        #qp.drawRect(xi3-5,yi3-5,10,10)
        W = 25 # 25 # 12.5

        xi4 = xi
        yi4 = yi
        xi5 = xf
        yi5 = yf

        # first quadrant (bottom right)
        if xf >= xi and yf >= yi and abs(xi-xf) >W:
            xi4 = xi+W
            yi4 = yi
            xi5 = xf-W
            yi5 = yf

        # second quadrant (bottom left)
        if xf < xi and yf >= yi and abs(xi-xf) >W:
            xi4 = xi-W
            yi4 = yi
            xi5 = xf+W
            yi5 = yf

        # third quadrant (top left)
        if xf < xi and yf < yi and abs(xi-xf) >W:
            xi4 = xi-W
            yi4 = yi
            xi5 = xf+W
            yi5 = yf

        # fourth quadrant (top right)
        if xf >= xi and yf < yi and abs(xi-xf) >W:
            xi4 = xi+W
            yi4 = yi
            xi5 = xf-W
            yi5 = yf

        xi6 = x0
        yi6 = y0
        xi7 = xi
        yi7 = yi

        # first quadrant (bottom right)
        if xi >= x0 and yi >= y0 and abs(xi-x0) >W:
            xi6 = x0+W
            yi6 = y0
            xi7 = xi-W
            yi7 = yi

        # second quadrant (bottom left)
        if xi < x0 and yi >= y0 and abs(xi-x0) >W:
            xi6 = x0-W
            yi6 = y0
            xi7 = xi+W
            yi7 = yi

        # third quadrant (top left)
        if xi < x0 and yi < y0 and abs(xi-x0) >W:
            xi6 = x0-W
            yi6 = y0
            xi7 = xi+W
            yi7 = yi

        # fourth quadrant (top right)
        if xi >= x0 and yi < y0 and abs(xi-x0) >W:
            xi6 = x0+W
            yi6 = y0
            xi7 = xi-W
            yi7 = yi

        # qp.drawRect(xi4-5,yi4-5,10,10)
        # qp.drawRect(xi5-5,yi5-5,10,10)
        # qp.drawRect(xi6-5,yi6-5,10,10)
        # qp.drawRect(xi7-5,yi7-5,10,10)

        xi62 = 0.5*(xi6+xi2)
        yi62 = 0.5*(yi6+yi2)

        """
        if self.darkMode:
            pen = QtGui.QPen(QtGui.QColor(50, 50, 50 , 255),0) # ,2, Qt.SolidLine)
        else:
            pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 255),0) # ,2, Qt.SolidLine)
        """
        #pen = QtGui.QPen(self.parent().theme_manager.get_color(1),0)
        #qp.setPen(pen)

        xi35 = 0.5*(xi3+xi5)
        yi35 = 0.5*(yi3+yi5)

        """
        qp.drawLine(x0,y0,xi6,yi6)
        qp.drawLine(xi6,yi6,xi62,yi62)
        qp.drawLine(xi62,yi62,xi2,yi2)
        qp.drawLine(xi2,yi2,xi,yi)
        qp.drawLine(xi,yi,xi3,yi3)

        qp.drawLine(xi3,yi3,xi3,yi3)
        qp.drawLine(xi3,yi3,xi35,yi35)
        qp.drawLine(xi35,yi35,xi5,yi5)
        qp.drawLine(xi5,yi5,xf,yf)
        """

        # https://stackoverflow.com/questions/52014197/how-to-interpolate-a-2d-curve-in-python
        # define some points
        end_is_start = False
        if abs(x0-xf) < 0.0002 and abs(y0-yf) < 0.0002:
            end_is_start = True
            points2 = np.array([[x0,xi62,xi2,xi,xd],
                                [y0,yi62,yi2,yi,yd]]).T
        else:
            # points2 = np.array([[x0,xi6,xi62,xi2,xi,xd,xi3,xi35,xi5,xf],
            #              [y0,yi6,yi62,yi2,yi,yd,yi3,yi35,yi5,yf]]).T

            points2 = np.array([[x0,xi,xd,xf],
                                 [y0,yi,yd,yf]]).T

            #points2 = np.array([[x0,xi2,xi,xd,xf],
            #                     [y0,yi2,yi,yd,yf]]).T

            # points2 = np.array([[x0,xi6,xi62,xi2,xi,xd,xi3,xi35,xi5,xf], [y0,yi6,yi62,yi2,yi,yd,yi3,yi35,yi5,yf]]).T

        points = points2 #  np.array(points2)
        assert len(points) > 0, "empty points"

        # points[0] = (self.raster(points[0][0]), self.raster(points[0][1]))
        # points = [(self.raster(t[0]), self.raster(t[1])) for t in points]

        xt,yt = 0,0
        # Linear length along the line:
        try:
            distance = np.cumsum( np.sqrt(np.sum( np.diff(points, axis=0)**2, axis=1 )) )
            distance = np.insert(distance, 0, 0)

            if len(distance) > 0:
                distance = distance/distance[-1]
            
            # Interpolation for different methods:
            # interpolations_methods = ['slinear', 'quadratic', 'cubic']

            if not self.show_raster:
                if not end_is_start:
                    # method = "quadratic"
                    method  = self.interpol_method
                    
                    # for method in interpolations_methods:
                    interpolated_points =  my_interpolate(points, method = method)
                    # interpolated_points = interpol(alpha)

                else:
                    # method  = self.interpol_method
                    # alpha = np.linspace(0, 1, 60) #np.linspace(0, 1, 80)

                    c0x = 0.5*(xi-x0)
                    c0y = 0.5*(yi-y0)

                    rx = max(3,0.5 * np.sqrt((x0-xi)**2 + (y0-yi)**2) )
                    ry = max(3,0.5 * np.sqrt((xi-xd)**2 + (yi-yd)**2) )

                    t0 = np.arctan2((yi-y0) ,(xi-x0))

                    t = t0 + np.linspace(0,2*np.pi, 100)

                    x_coords = [rx * np.cos(ti) * np.cos(t0) - ry * np.sin(ti) * np.sin(t0) + c0x + x0 for ti in t]
                    y_coords = [rx * np.cos(ti) * np.sin(t0) + ry* np.sin(ti) * np.cos(t0) + c0y + y0 for ti in t]

                    interpolated_points = list(zip(x_coords,y_coords))


            else:
                method = "linear"
                #alpha = np.linspace(0, 1, 4) # np.linspace(0, 1, 50)
                #for method in interpolations_methods:

                #for method in interpolations_methods:
                #interpol_x =  interp1d(distance, [p[0] for p in points], kind=method, axis=0)
                #interpol_y =  interp1d(distance, [p[1] for p in points], kind=method, axis=0)
                #interpolated_points_x = interpol_x(alpha)
                #interpolated_points_y = interpol_y(alpha)
                #n = len(interpolated_points_y)
                #interpolated_points = [(interpolated_points_x[i], interpolated_points_y[i]) for i in range(n)]
                interpolated_points =  my_interpolate(points, method = method, n=10)
                # print("points", points)

                x0 = (interpolated_points[0][0])
                y0 = (interpolated_points[0][1])
                xf = (interpolated_points[-1][0])
                yf = (interpolated_points[-1][1])
                
                # start_is_end = abs(x0 - xf) < 1e-6 and abs(y0 - yf) < 1e-6

                if not end_is_start:
                    
                    # construct a line
                    xc1, yc1 = self.manhattan_lines(x0,y0,xi,yi,111)
                    xc2, yc2 = self.manhattan_lines(xi,yi,xd,yd,111)
                    xc3, yc3 = self.manhattan_lines(xd,yd,xf,yf,111)

                    x_coords = xc1 + xc2 + xc3
                    y_coords = yc1 + yc2 + yc3
                else:
                    # construct a circle with center point x0, y0
                    # through xi,yi

                    c0x = 0.5*(xi-x0)
                    c0y = 0.5*(yi-y0)

                    rx = max(3,0.5 * np.sqrt((x0-xi)**2 + (y0-yi)**2) )
                    ry = max(3,0.5 * np.sqrt((xi-xd)**2 + (yi-yd)**2) )

                    t0 = np.arctan2((yi-y0) ,(xi-x0))

                    t = t0 + np.linspace(0,2*np.pi, 100)

                    x_coords = [rx * np.cos(ti) * np.cos(t0) - ry * np.sin(ti) * np.sin(t0) + c0x + x0 for ti in t]
                    y_coords = [rx * np.cos(ti) * np.sin(t0) + ry* np.sin(ti) * np.cos(t0) + c0y + y0 for ti in t]      

                interpolated_points = list(zip(x_coords,y_coords))


            if self.show_raster:
                interpolated_points = [(self.raster(t[0]), self.raster(t[1])) for t in interpolated_points]

            # draw connector line as path
            n = len(interpolated_points)

            # print("interpolated_poitns",interpolated_points)
            # draw curved line
            offset_text = 20

            qp.setRenderHint(QPainter.Antialiasing)

            connPath = QPainterPath()
            startx = self.zoom_factor*(interpolated_points[0][0])
            starty = self.zoom_factor*(interpolated_points[0][1])

            connPath.moveTo(startx,starty)

            for i in range(1,int(n/2)-offset_text):
                #qp.drawLine(self.zoom_factor*(interpolated_points[i-1][0]),self.zoom_factor*(interpolated_points[i-1][1]),
                #            self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))
                connPath.lineTo(self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))

            for i in range(int(n/2)-offset_text,len(interpolated_points)):
                #qp.drawLine(self.zoom_factor*(interpolated_points[i-1][0]),self.zoom_factor*(interpolated_points[i-1][1]),
                #            self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))
                connPath.lineTo(self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))

            connPath.moveTo(startx,starty)
            connPath.closeSubpath()
            qp.drawPath(connPath)

            
            # draw arrow at midpoint
            md = int(n/2.)
            # rint("md", md)
            # interpolated_points = [(self.raster(p[0],50),self.raster(p[1],50)) for p in interpolated_points]
            draw_reverse = False

            if not end_is_start:
                xt, yt = self.draw_arrow_at(qp,interpolated_points,md,dashed=dashed)

            if not is_unidir: # double arrow for non-unidirectional
                if not end_is_start and self.main_widget.checkBox.isChecked():
                    if not draw_reverse:

                        if self.main_widget.checkBox_raster.isChecked():
                            md = int(n/2.)
                        else:
                            md = int(n/2. - 15)

                        xt, yt = self.draw_arrow_at(qp, interpolated_points,md, reverse=True,dashed=dashed)


            # interpolated_points = [(self.raster(t[0]), self.raster(t[1])) for t in points]


        except Exception as e:
            print("Error in Line Interpolation:",str(e))
            traceback.print_exc()
            xt = 0
            yt = 0

        """
        if self.darkMode:
            pen = QtGui.QPen(QtGui.QColor(255, 255, 255 , 255),0) # ,2, Qt.SolidLine)
        else:
            pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 255),0) # ,2, Qt.SolidLine
        """

        #pen = QtGui.QPen(self.parent().theme_manager.get_color(3),0)
        #qp.setPen(pen)

        # points are [y0,xi6,xi2,xi3,xi5,xf]

        # qp.drawLine(xi2,yi2,xi5,yi5)

        #qp.drawLine(xi2,yi2,xi7,yi7)
        #qp.drawLine(xi7,yi7,xf,yf)

        #qp.drawLine(xi2,yi2,xf,yf)

        #self.draw_bezier(qp,xi4,yi4,xi5,yi5,xi3,yi3) # starting point
        #self.draw_bezier(qp,xi7,yi7,xi6,yi6,xi2,yi2) # starting point
        #self.draw_bezier(qp,xi,yi,xi3,yi3,xi7,yi7) # starting point
        #self.draw_bezier(qp,xi3,yi3,xf,yf,xi5,yi5) # starting point

        # return position for label
        return xt,yt

    def dawPreview(self,qp):
        # draw small preview window
        # param qp: a qpainter object
        if not self.main_widget.checkBox_9.isChecked():
            return

        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1)
        qp.setPen(pen)

        # draw Origin
        #qp.drawLine(self.zoom_factor*(0.5*self.W0-5+self.gox),self.zoom_factor*(0.5*self.H0+self.goy),self.zoom_factor*(0.5*self.W0+5+self.gox),self.zoom_factor*(0.5*self.H0+self.goy))
        #qp.drawLine(self.zoom_factor*(0.5*self.W0+self.gox),self.zoom_factor*(0.5*self.H0-5+self.goy),self.zoom_factor*(0.5*self.W0+self.gox),self.zoom_factor*(0.5*self.H0+5+self.goy))

        w_frame = 80
        h_frame = 80
        distance = 25
        move_x = 10 # self.W/2 #  distance + 10
        move_y = 10 # -25

        brush = QtGui.QBrush(QtGui.QColor(80,80,80 , 100),1) # QtGui.QBrush(self.parent().theme_manager.get_color(3)) #7
        qp.fillRect(QtCore.QRectF(move_x-5,move_y-5,w_frame+10,h_frame+10),brush)

        if len(self.boxes) == 0:
            return

        x_min = min(0,-self.gox)
        x_max = max(self.W,self.W-self.gox)
        y_min = min(0,-self.goy)
        y_max = max(self.H,self.H-self.goy)
        x_vals = []
        y_vals = []

        for box in self.boxes:
            x_vals.append(box.x)
            y_vals.append(box.y)


        x_max = max(np.max(x_vals),-self.gox)
        y_max = max(np.max(y_vals),-self.goy)
        x_min = min(np.min(x_vals),-self.gox)
        y_min = min(np.min(y_vals),-self.goy)

        if x_max == x_min:
            x_max = x_min + 10
        if y_max == y_min:
            y_max = y_min + 10

        if x_max < self.W:
            x_max = self.W
        if y_max < self.H:
            y_max = self.H

        x_vals = [self.zoom_factor*(xi+self.gox)/10 for xi in x_vals]
        y_vals = [self.zoom_factor*(yi+self.goy)/10 for yi in y_vals]

        # origin
        x_vals.append(self.zoom_factor/10 *(self.gox + 0.5*self.W))
        y_vals.append(self.zoom_factor/10 *(self.goy + 0.5*self.H))

        # pointer
        x_vals.append(1/10 *(0.5*self.W))
        y_vals.append(1/10 *(0.5*self.H))

        mx = np.mean(x_vals)
        my = np.mean(y_vals)

        x_vals = [x-mx for x in x_vals]
        y_vals = [y-my for y in y_vals]

        dx = max(-np.min(x_vals),np.max(x_vals))
        dy = max(-np.min(y_vals),np.max(y_vals))

        x_vals = [0.5*w_frame*x/dx for x in x_vals]
        y_vals = [0.5*h_frame*y/dy for y in y_vals]

        x_vals = [x+0.5*w_frame for x in x_vals]
        y_vals = [y+0.5*h_frame for y in y_vals]

        x_vals = [move_x +int(distance/2)+x-2-10 for x in x_vals]
        y_vals = [move_y +int(distance/2)+y-2-10 for y in y_vals]

        # transform all points

        for i in range(len(x_vals)-2):
            x = int(x_vals[i])
            y = int(y_vals[i])
            if self.boxes[i].ishelper:
                    qp.drawRect(x,y,1,1)
            else:
                #if self.boxes[i] == self.highlighted:
                #    pen = QtGui.QPen(self.parent().theme_manager.get_color(3),1)
                #    brush = QtGui.QBrush(self.parent().theme_manager.get_color(3))
                #    qp.setPen(pen)

                #    qp.fillRect(QtCore.QRectF(x,y,4,4),brush)
                #else:
                qp.drawRect(x,y,4,4)
        
        # skip origin ...
        x = x_vals[-1]
        y = y_vals[-1]
        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(4),1)
        qp.setPen(pen)
        ox = x
        oy =  y
        d = 5
        qp.drawLine(int(ox-d),int(oy),int(ox+d),int(oy))
        qp.drawLine(int(ox),int(oy-d),int(ox),int(oy+d))

        # draw Origin
        # ox = (self.gox-x_min)/(x_max-x_min)*(w_frame-distance)
        # oy = (self.goy-y_min)/(y_max-y_min)*(h_frame-distance)
        # dx = 5
        # dy = 5
        # ox_transformed = move_x + self.W-w_frame+ox-2
        # oy_transformed = move_y + self.H-h_frame+oy-2

        # qp.drawLine(ox_transformed-dx,oy_transformed,ox_transformed+dx,oy_transformed)
        # qp.drawLine(ox_transformed,oy_transformed-dy,ox_transformed,oy_transformed+dy)

    def raster(self, x, n=19):
        if self.show_raster: # show rasterized
            # x_rastered = np.round(x,int(-np.log10(self.raster_size)))
            m = min(n,self.raster_size)
            x_rastered = (int(1.0 * x/m))*m  # np.round(x,int(-np.log10(self.raster_size)))
            return int(x_rastered)
        return int(x)

    def paint(self,reference,ratio=1):

        qp = QtGui.QPainter(reference)
        qp.scale(ratio, ratio)
        #qp.setRenderHint(QtGui.QPainter.SmoothPixmapTransform )
        qp.setRenderHint(QtGui.QPainter.HighQualityAntialiasing  )
        qp.setRenderHint(QtGui.QPainter.TextAntialiasing  )


        for box in self.boxes:
            box.x = self.raster(box.x)
            box.y = self.raster(box.y)

        """
        if self.darkMode:
            pen = QtGui.QPen(QtGui.QColor(120, 120, 120, 255),1) # ,2, Qt.SolidLine)
        else:
            pen = QtGui.QPen(QtGui.QColor(0, 0, 0, 255),1) # ,2, Qt.SolidLine)
        """
        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1)
        qp.setPen(pen)


        if self.main_widget.checkBox_7.isChecked(): # draw boundary rect with white background?
            br_white = QtGui.QBrush(QtGui.QColor(255, 255, 255, 255),1)
            qp.fillRect(QtCore.QRectF(QtCore.QPointF(0,0),QtCore.QPointF(int(self.W-2),int(self.H-2))),br_white) # background
            if reference == self:
                qp.drawRect(0,0,self.W-2,self.H-4)
        else:
            if reference == self:
                qp.drawRect(0,0,self.W-2,self.H-4)


        # draw grid 
        if self.main_widget.checkBox_raster_2.isChecked():
            for i in np.arange(0, int(self.W-2), self.raster_size):
                for j in np.arange(0, int(self.H-2),self.raster_size):
                    qp.drawRect(int(i)-1,int(j)-1,2,2)

        # pen2 = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
        # qp.setPen(pen2)
        # qp.drawRect(0+self.gox,0+self.goy,self.gox + self.W-2,self.goy + self.H-2)

        qp.setPen(pen)

        filter_vals = set([item.text() for item in self.main_widget.filterCombo.selectedItems()])
        active_boxes = []
        active_connectors = []

        for conn in self.connectors:
            no_selection = len(filter_vals) == 0
            is_selected = len(filter_vals) != 0 and len(list(set(conn.items) & filter_vals)) > 0
            if no_selection or is_selected:
                if conn not in active_connectors:

                    if self.highlighted is None:
                        active_connectors.append(conn)
                    else:
                        if conn.a == self.highlighted or conn.b == self.highlighted or conn.c == self.highlighted:
                            active_connectors.append(conn)


        for conn in self.connectors:

            is_unidir = True
            is_flow = False

            if conn.c == self.highlighted:

                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(4),2.0) # ,2, Qt.SolidLine)

                is_unidir = True
                for i,data in enumerate(self.main_widget.entry_data): # TODO make this more efficient

                    # print("Connector data", data)
                    if data["shortname"] == conn.c.name:

                        if str(data["uni-directional"]) == "False":
                            is_unidir = False

                        break
            else:
                # pen = QtGui.QPen(QtGui.QColor(0, 0, 0, 200),1) # ,2, Qt.SolidLine)

                """
                if self.darkMode:
                    pen = QtGui.QPen(QtGui.QColor(120, 120, 120, 255),1) # ,2, Qt.SolidLine)
                else:
                    pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 75),0) # ,2, Qt.SolidLine)
                """

                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1.5)

                if conn in active_connectors:

                    # is actually a flow?
                    is_flow = False
                    is_unidir = True

                    for i, data in enumerate(self.main_widget.entry_data): # TODO make this more efficient
                        # print("connector data", data)
                        if data["shortname"] == conn.c.name:

                            if str(data["log transaction"]) == "True":
                                is_flow = True

                            if str(data["uni-directional"]) == "False":
                                is_unidir = False

                            break;

                    if is_flow:
                        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1.5)
                    else:
                        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(4),1.5)

                    pen.setStyle(Qt.SolidLine)
                    pen.setWidthF(1.5)

                else:
                    pen.setStyle(Qt.DashLine)
                    #pen.setDashOffset(4.0)
                    pen.setDashPattern([1,4])
                    pen.setWidthF(1.1)

                    is_flow = False
                    is_unidir = True


            qp.setPen(pen)

            # qp.drawLine(int(conn.a.x+25), int(conn.a.y+25), int(conn.c.x+5), int(conn.c.y+5))
            # qp.drawLine(int(conn.c.x+5), int(conn.c.y+5), int(conn.b.x+25), int(conn.b.y+25))

            # apply filter: does the connector contain any of the filter_vals?

            show = True
            offset_box = 25

            if conn in active_connectors:
                # draw only if filter applies

                tx,ty = self.curved_connector(qp,self.gox + int(conn.a.x+offset_box),
                                                 self.goy + int(conn.a.y+offset_box),
                                                 self.gox + int(conn.b.x+offset_box),
                                                 self.goy + int(conn.b.y+offset_box),
                                                 # self.gox + int(conn.c.x+5), self.goy + int(conn.c.y+5),
                                                 # self.gox + int(conn.d.x+5), self.goy + int(conn.d.y+5),
                                                 self.gox + int(conn.c.x), self.goy + int(conn.c.y),
                                                 self.gox + int(conn.d.x), self.goy + int(conn.d.y),
                                                 is_flow=is_flow,
                                                 is_unidir=is_unidir,
                                              )


                # self.curved_connector(qp,int(conn.a.x), int(conn.a.y),int(conn.b.x), int(conn.b.y),int(conn.c.x), int(conn.c.y))
                if conn.a not in active_boxes:
                    active_boxes.append(conn.a)
                if conn.b not in active_boxes:
                    active_boxes.append(conn.b)
                if conn.c not in active_boxes:
                    active_boxes.append(conn.c)
                if conn.d not in active_boxes:
                    active_boxes.append(conn.d)

                if conn.c == self.highlighted:
                    pen = QtGui.QPen(self.main_widget.theme_manager.get_color(6),1.5)
                    qp.setPen(pen)
                else:
                    pen = QtGui.QPen(self.main_widget.theme_manager.get_color(1),1.5)
                    qp.setPen(pen)


                if self.main_widget.checkBox_5.isChecked(): # draw at handles
                    if self.main_widget.checkBox_2.isChecked():


                        if self.highlighted == conn.c:
                            brush = QtGui.QBrush(self.main_widget.theme_manager.get_color(4),1)
                        else:
                            brush = QtGui.QBrush(self.main_widget.theme_manager.get_color(3),1)

                        fontsize = 8

                        try:
                            fontsize = int(self.main_widget.spinBoxFontSize_2.value())
                            qp.setFont(QFont("Arial", fontsize))

                        except Exception as e:
                            print(str(e))

                        conn.c.label.draw(qp,brush)

                        """
                        a = self.zoom_factor*(self.gox +conn.c.x)
                        b = self.zoom_factor*(self.goy + conn.c.y-7)

                        # draw second connector
                        a2 = self.zoom_factor*(self.gox +conn.d.x)
                        b2 = self.zoom_factor*(self.goy + conn.d.y-7)

                        # draw connector label


                        if self.main_widget.checkBox_4.isChecked():

                            fm = qp.fontMetrics()
                            text_length = fm.boundingRect(conn_label_name).width() + 2
                            text_height = fm.boundingRect(conn_label_name).height() + 2

                            # text_length = 5.1*len(conn.c.name)
                            qp.fillRect(QtCore.QRectF(QtCore.QPointF(int(a-.1),int(b-text_height+1)),QtCore.QPointF(int(a+text_length),int(b))),brush)
                            # qp.fillRect(QtCore.QRectF(QtCore.QPointF(int(a2-.1),b2-10),QtCore.QPointF(int(a2+5.1*len(conn.d.name)),b2+10)),brush)

                        qp.drawText(int(a),int(b)-1,conn_label_name)
                        """

                else: # do not draw at handles
                    if self.main_widget.checkBox_2.isChecked():
                        if self.highlighted == conn.c:
                            brush = QtGui.QBrush(self.main_widget.theme_manager.get_color(4),1)
                        else:
                            brush = QtGui.QBrush(self.main_widget.theme_manager.get_color(3),1)


                        if conn.a.x != conn.b.x and conn.a.y != conn.b.y:
                            # label ends and starts at different box

                            a = self.zoom_factor*(tx)
                            b = self.zoom_factor*(ty)

                            a2 = self.zoom_factor*(tx+5)
                            b2 = self.zoom_factor*(ty+5)


                        else: # label ends and starts at same box
                            a = self.zoom_factor*(self.gox +conn.c.x)
                            b = self.zoom_factor*(self.goy + conn.c.y)

                            a2 = self.zoom_factor*(self.gox +conn.d.label.x + conn.d.label.offx)
                            b2 = self.zoom_factor*(self.goy + conn.d.label.y+conn.d.label.offy)


                        fontsize = 8

                        try:
                            fontsize = int(self.main_widget.spinBoxFontSize_2.value())
                            qp.setFont(QFont("Arial", fontsize))

                        except Exception as e:
                            print(str(e))

                        conn.c.label.draw(qp, brush, position = (a2,b2))

                        """
                        DEPRICATED


                        # draw agent labels


                        if self.main_widget.checkBox_4.isChecked():
                            # text_length = 5.1*len(conn.c.name)

                            fm = qp.fontMetrics()
                            text_length = fm.boundingRect(conn.c.name.replace("_"," ")).width() + 2
                            text_height = fm.boundingRect(conn.c.name.replace("_"," ")).height() + 2

                            qp.fillRect(QtCore.QRectF(QtCore.QPointF(int(a-.1),int(b-text_height+1)),QtCore.QPointF(int(a+text_length),int(b))),brush)
                            # qp.fillRect(QtCore.QRectF(QtCore.QPointF(int(a2-.1),b2-10),QtCore.QPointF(int(a2+5.1*len(conn.d.name)),b2+10)),brush)


                        qp.drawText(int(a),int(b)-1,conn_label_name)

                        # draw second connector
                        """

            else:

                """
                self.curved_connector(qp,self.gox + int(conn.a.x+12.5),
                 self.goy + int(conn.a.y+12.5),
                 self.gox + int(conn.b.x+12.5),
                 self.goy + int(conn.b.y+12.5),
                 self.gox + int(conn.c.x+5),
                 self.goy + int(conn.c.y+5),
                 self.gox + int(conn.d.x+5),
                 self.goy + int(conn.d.y+5),
                 dashed=True,
                 is_flow=is_flow)
                """
                tx,ty = self.curved_connector(qp,self.gox + int(conn.a.x+offset_box),
                                                 self.goy + int(conn.a.y+offset_box),
                                                 self.gox + int(conn.b.x+offset_box),
                                                 self.goy + int(conn.b.y+offset_box),
                                                 self.gox + int(conn.c.x+5), self.goy + int(conn.c.y+5),
                                                 self.gox + int(conn.d.x+5), self.goy + int(conn.d.y+5),
                                                 is_flow=is_flow,
                                                 is_unidir=is_unidir
                                              )


        """
        if self.darkMode:
            pen = QtGui.QPen(QtGui.QColor(120, 120, 120, 255),1) # ,2, Qt.SolidLine)
        else:
            pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 75),0) # ,2, Qt.SolidLine)
        """
        pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1)
        pen.setStyle(Qt.SolidLine)

        # pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 200),1) # ,2, Qt.SolidLine)
        qp.setPen(pen)

        # print("self.boxes",[b.name for b in self.boxes])

        """
        FOR DEBUGGING LABELS
        for box in self.boxes:
            if box.name != "":
                z = self.zoom_factor
                a = z*(box.label.x + box.label.offx +self.gox )
                b = z*(box.label.y + box.label.offy -11+self.goy ) # self.zoom_factor*(y-7+self.goy)

                w = z* 25
                w2 = z* 15
                qp.drawRect(QtCore.QRectF(a,b,w,w2))
        """

        for box in self.boxes:

            #box.x = self.raster(box.x)
            #box.y = self.raster(box.y)

            if box != self.selected: # and box != self.connected2:

                x = box.x
                y = box.y

            elif box == self.selected:

                x = self.mousex + self.offx
                y = self.mousey + self.offy

                box.x = x
                box.y = y

                #x = box.x * (1-self.zoom_factor) + new_x *(self.zoom_factor)
                #y = box.y * (1-self.zoom_factor) + new_y *(self.zoom_factor)
                #box.label.x = x
                #box.label.y = y

            #x = box.x
            #y = box.y

            if box == self.selected:
                br = QtGui.QBrush(QtGui.QColor(34, 34, 230, 255))
            elif box == self.connected1:
                br = QtGui.QBrush(QtGui.QColor(100, 10, 10, 255))
            elif box.ishelper and box != self.highlighted:

                """
                if self.darkMode:
                    br = QtGui.QBrush(QtGui.QColor(255, 255, 255, 255))
                else:
                    br = QtGui.QBrush(QtGui.QColor(74, 138, 217, 255))
                """
                br = QtGui.QBrush(self.main_widget.theme_manager.get_color(0),1)


            elif box == self.highlighted:
                br = QtGui.QBrush(self.main_widget.theme_manager.get_color(5))
            else:
                """
                if self.darkMode:
                    br = QtGui.QBrush(QtGui.QColor(20, 20, 20, 255))
                else:
                    br = QtGui.QBrush(QtGui.QColor(200, 200, 200, 255))
                """
                br = QtGui.QBrush(self.main_widget.theme_manager.get_color(0),1)

            x = int(x)
            y = int(y)

            if box.ishelper:
                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(6),1)

                if box.is_support:
                    br = QtGui.QBrush(self.main_widget.theme_manager.get_color(7),1)
                else:
                    br = QtGui.QBrush(self.main_widget.theme_manager.get_color(1),1)

                qp.setPen(pen)
                # qp.drawText(x,y-7,box.name)

            else:
                #qp.drawText(x,y-3,box.name.capitalize())

                # draw agent labels


                #a = self.zoom_factor*(x+self.gox)
                #b = self.zoom_factor*(y-11+self.goy) # self.zoom_factor*(y-7+self.goy)
                a = self.zoom_factor*(x+self.gox )
                b = self.zoom_factor*(y-11+self.goy ) # self.zoom_factor*(y-7+self.goy)

                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(6),1)
                qp.setPen(pen)


                # draw agent labels (box labels)
                fontsize = 12

                try:
                    fontsize = int(self.main_widget.spinBoxFontSize.value())
                    qp.setFont(QFont("Arial", fontsize))

                except Exception as e:
                    print(str(e))



                if self.main_widget.checkBox_4.isChecked():

                    br_label = QtGui.QBrush(self.main_widget.theme_manager.get_color(3),1)
                else:
                    br_label = None

                """
                    fm = qp.fontMetrics()
                    text_length = fm.boundingRect(box.name.replace("_"," ")).width() + 2
                    text_height = fm.boundingRect(box.name.replace("_"," ")).height() + 2

                    qp.fillRect(QtCore.QRectF(QtCore.QPointF(int(a-.1),int(b-text_height + 1)),QtCore.QPointF(int(a+text_length),int(b))),br_label) # background


                qp.drawText(int(a) ,int(b)-1, box.name) # text
                """

                box.label.draw(qp,br_label)


            qp.setBrush(br)

            if box.ishelper:

                x2 = x + 6
                y2 = y + 6

                pen = QtGui.QPen(QtGui.QColor(0, 0, 0 , 0),0) # ,2, Qt.SolidLine)
                qp.setPen(pen)
                # qp.drawRect(QtCore.QRectF(QtCore.QPointF(x,y),QtCore.QPointF(x2,y2)))

                if box.is_support:
                    show_box = self.main_widget.checkBox_6.isChecked()
                else:
                    show_box = self.main_widget.checkBox_3.isChecked()

                if show_box:
                    if box in active_boxes:

                        """
                        if box.is_support:
                            qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),6, 6)
                        else:
                            qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),6, 6)
                        """
                        box.draw(qp,active=True)

                    else:

                        """
                        if box.is_support:
                            qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),2, 2)
                        else:
                            qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),2, 2)
                        """
                        box.draw(qp,active=False)

                """
                if self.darkMode:
                    pen = QtGui.QPen(QtGui.QColor(120, 120, 120 , 200),0) # ,2, Qt.SolidLine
                else:
                    pen = QtGui.QPen(QtGui.QColor(0, 0, 0 , 200),0) # ,2, Qt.SolidLine)
                """
                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(2),1)
                qp.setPen(pen)

            else:

                #W = 45

                #x2 = x + W
                #y2 = y + W

                pen = QtGui.QPen(QtGui.QColor(0, 0, 100 , 0),0) # ,2, Qt.SolidLine)
                #br = QtGui.QBrush(QtGui.QColor(0, 0, 100, 255))
                qp.setPen(pen)
                # qp.setBrush(br)

                """
                qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox + x+W-5),self.zoom_factor*(self.goy + y+W-5)),11, 10)
                qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox +x+W-5),self.zoom_factor*(self.goy +y+5)),11, 10)
                qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox +x+5),self.zoom_factor*(self.goy +y+W-5)),11,10)
                qp.drawEllipse(QtCore.QPointF(self.zoom_factor*(self.gox +x+5),self.zoom_factor*(self.goy +y+5)),11, 10)

                qp.drawRect(QtCore.QRectF(QtCore.QPointF(self.zoom_factor*(self.gox +x+5),self.zoom_factor*(self.goy +y-5)),QtCore.QPointF(self.zoom_factor*(self.gox +x2-5),self.zoom_factor*(self.goy +y+5))))
                qp.drawRect(QtCore.QRectF(QtCore.QPointF(self.zoom_factor*(self.gox +x+5),self.zoom_factor*(self.goy +y+W-5)),QtCore.QPointF(self.zoom_factor*(self.gox +x2-5),self.zoom_factor*(self.goy +y+W+5))))

                qp.drawRect(QtCore.QRectF(QtCore.QPointF(self.zoom_factor*(self.gox +x-6),self.zoom_factor*(self.goy +y+1)),QtCore.QPointF(self.zoom_factor*(self.gox +x+5),self.zoom_factor*(self.goy +y+W-5))))
                qp.drawRect(QtCore.QRectF(QtCore.QPointF(self.zoom_factor*(self.gox +x+W-5),self.zoom_factor*(self.goy +y+1)),QtCore.QPointF(self.zoom_factor*(self.gox +x+W+5),self.zoom_factor*(self.goy +y+W-5))))
                """


                if box in active_boxes:
                    box.draw(qp,brush=br,active=True)
                else:
                    box.draw(qp,brush=br,active=False)




                """
                if self.darkMode:
                    pen = QtGui.QPen(QtGui.QColor(120,120,120 , 255),0) # ,2, Qt.SolidLine)
                else:
                    pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 255),0) # ,2, Qt.SolidLine)
                """

                pen = QtGui.QPen(self.main_widget.theme_manager.get_color(3),1)
                qp.setPen(pen)
        self.dawPreview(qp)




        return qp

    def paintEvent(self, event):
        self.paint(self)


    def hit_box(self,x,y,xb,yb,wx,wy):
        """
        check if (x,y) lies within the rectangle
        (xb, yb), (xb, yb+wy), ...., (xb+wx, yb+wy)
        """
        if wx < 30:
            cond1 = self.zoom_factor*(xb+self.gox - wx)  < x*self.zoom_factor  < self.zoom_factor*(xb+ self.gox + wx)
            cond2 = self.zoom_factor*(yb+self.goy - wy) < y*self.zoom_factor  < self.zoom_factor*(yb + wy + self.goy)
        else:
            cond1 = self.zoom_factor*(xb+self.gox)  < x*self.zoom_factor  < self.zoom_factor*(xb+ self.gox + wx)
            cond2 = self.zoom_factor*(yb+self.goy) < y*self.zoom_factor  < self.zoom_factor*(yb + wy + self.goy)
        return cond1 and cond2

    def mousePressEvent(self, event):
        # print("MOUSE PRESS DRAW WIDGET")

        self.main_widget.update_graphics_data(self.boxes,self.connectors)

        time_delay = time.time() - self.last_click
        self.last_click = time.time()


        if event.buttons() & QtCore.Qt.RightButton:
            self.mode = "drag"

        elif event.buttons() & QtCore.Qt.LeftButton:
            self.mode = "select"

        x,y = event.pos().x(), event.pos().y()

        x = int(x/self.zoom_factor)
        y = int(y/self.zoom_factor)

        self.go_startx = x
        self.go_starty = y
        self.gox_start = self.gox
        self.goy_start = self.goy

        # check if any box was selected
        box_hit = False
        self.selected = None

        # check for labels
        for box in self.boxes:

            if self.mode != "move" and self.hit_box(x, y, box.label.x + box.label.offx, box.label.y + box.label.offy -11, 25, 15):
                self.mode = "drag_label"
                self.selected = box.label
                print("selected", box.name)

        # check for box movements
        for box in self.boxes:
            if box.ishelper:
                w = 10
            else:
                w = 50

            if self.mode != "move" and self.hit_box(x, y, box.x, box.y, w, w):
                # print("box hit", box.name)
                # self.mode = "drag"

                box_hit = True
                if self.mode == "drag":
                    self.selected = box
                    self.offx = (box.x - x)
                    self.offy = (box.y - y)

                elif self.mode == "select":
                    if time_delay < 0.2: # double click on agent to edit
                        if self.highlighted is None:
                            self.highlighted = box
                        self.edit_agent()
                        self.highlighted = None
                        return
                    else:
                        if self.highlighted == box:
                            self.highlighted = None
                            #print("unhighlight", box)
                            self.update()
                        else:
                            #print("highlight",box)
                            self.highlighted = box

                    # self.mode = "drag"

                    if box.ishelper and self.highlighted == box:
                        self.main_widget.data_select_where(box.name)

                    self.update()

                elif self.mode == "edit":
                    #print("EDIT", box)

                    if box.ishelper:
                        #print("EDIT CONNECTOR")
                        new_editor = TransactionEditor(self,box)

                    else:
                        #print("EDIT AGENT")
                        new_editor = AgentEditor(self,box)

                    self.mode = "select"

                elif self.mode == "delete":
                    raise RuntimeError("Delete mode not supported")
                    self.selected = None
                    self.connected1 = None
                    self.connected2 = None

                    if box.ishelper and box.parent is not None:
                        self.boxes.remove(box.parent.c)
                        self.boxes.remove(box.parent.d)
                        self.connectors.remove(box.parent)
                        self.update()
                        self.mode = "drag"
                    else:


                        for conn in self.connectors:
                            if conn.a == box or conn.b == box:
                                self.connectors.remove(conn)
                                self.boxes.remove(conn.a)
                                self.boxes.remove(conn.b)
                                self.boxes.remove(conn.c)
                                self.boxes.remove(conn.d)

                        if box in self.boxes:
                            self.boxes.remove(box)

                        self.update()
                        self.mode= "drag"


                elif self.mode == "connect":
                    self.selected = None

                    if self.connected1 is None:
                        self.connected1 = box
                        self.update()

                    elif self.connected2 is None and not box.ishelper:
                        self.connected2 = box

                        con = Connector(self.connected1,self.connected2)
                        self.connectors.append(con)
                        self.boxes.append(con.c)
                        self.boxes.append(con.d)

                        self.connected1 = None
                        self.connected2 = None
                        self.mode = "drag"

        if not box_hit and event.buttons() & QtCore.Qt.LeftButton:
            #print("Move")
            self.mode = "move"

    def mouseMoveEvent(self,event):

        x,y = event.pos().x(), event.pos().y()

        x = int(x/self.zoom_factor)
        y = int(y/self.zoom_factor)

        self.mousex = x
        self.mousey = y

        if self.mode == "drag_label":

            dx = x - self.go_startx
            dy = y - self.go_starty

            self.selected.offx = dx # max(-25,min(25,dx))
            self.selected.offy = dy # max(-25,min(25,dy))

        elif self.mode == "move":
            x = x
            y = y

            dx = x  - self.go_startx
            dy = y - self.go_starty
            # print("move", dx,dy)

            self.gox = self.raster(self.gox_start + dx)
            self.goy = self.raster(self.goy_start + dy)

        # print("move",x,y)
        else:

            hovering = False
            for box in self.boxes:
                
                if box.ishelper:
                    w = 20
                    cond = self.zoom_factor*(box.x+self.gox - w)  < x*self.zoom_factor  < self.zoom_factor*(box.x+self.gox + w) and self.zoom_factor*(box.y+self.goy - w) < y*self.zoom_factor  < self.zoom_factor*(box.y + w+self.goy)
                else:
                    w = 50
                    cond = self.zoom_factor*(box.x+self.gox)  < x*self.zoom_factor  < self.zoom_factor*(box.x+self.gox + w) and self.zoom_factor*(box.y+self.goy ) < y*self.zoom_factor  < self.zoom_factor*(box.y + w+self.goy)

                if self.mode != "move" and cond :

                    if event.buttons() & QtCore.Qt.RightButton:
                        self.mode = "drag"

                    self.setCursor(Qt.OpenHandCursor)
                    hovering = True

            if not hovering:
                self.setCursor(Qt.ArrowCursor)
                #self.parent().setCursor(Qt.OpenHandCursor)

        self.update()

    def mouseReleaseEvent(self, event):

        self.gox_start = self.gox
        self.goy_start = self.goy
        self.go_startx = self.gox
        self.go_starty = self.goy

        if self.mode == "drag_label":

            if self.selected is not None:
                self.selected.x += self.selected.offx
                self.selected.y += self.selected.offy

                self.selected.offx = 0
                self.selected.offy = 0

                self.old_positions[self.selected.name+"<label>"] = (self.selected.x, self.selected.y)
                self.label_position_data[self.selected.name+"<label>"] = {"x":self.selected.x, "y":self.selected.y}
                print("drag label",  (self.selected.x, self.selected.y))
                print(self.old_positions)

        if self.mode == "drag":
            x, y = event.pos().x(), event.pos().y()

            # m = self.raster_size
            x = int(x/self.zoom_factor) # self.raster(int(x/self.zoom_factor),m)
            y = int(y/self.zoom_factor) # self.raster(int(y/self.zoom_factor),m)

            self.mousex = x
            self.mousey = y

            if self.selected is not None:

                self.selected.x = self.mousex + self.offx # np.clip(self.mousex + self.offx,1-self.gox,2*self.maxx-self.gox)
                self.selected.y = self.mousey + self.offy # np.clip(self.mousey + self.offy, 1-self.goy,2*self.maxy-self.goy)
                self.selected.x0 = self.selected.x
                self.selected.y0 = self.selected.y

                dx = self.selected 
                
                self.old_positions[self.selected.name] = (self.selected.x, self.selected.y)
                try:
                    self.old_positions[self.selected.label.name + "<label>"] = (self.selected.label.x, self.selected.label.y)
                    self.label_position_data[self.selected.name+"<label>"] = {"x":self.selected.label.x, "y":self.selected.label.y}
                except:
                    pass 
                
                print("MOVE BOX", self.selected.name, (self.selected.x, self.selected.y))
                print("old positions", self.old_positions)
                

                self.selected = None
            self.update()

        self.mode = "select"



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


from ..core.clock import Clock
import warnings


class Slot:
    """
    Slot is the receiver of an information stream
    """
    _instances = {}

    def __init__(self, name, asynchronous = False):
        """
        This is a slot.
        :param name: name of the slot
        :param asynchronous:
        """

        self.name = name
        self.data = {}
        self.asynchronous = asynchronous
        self.__class__._instances[self.name] = self

    def __repr__(self):
        return "[Oo. SLOT %s]" % self.name

    @classmethod
    def retrieve(cls, name):
        """
        find a certain slot by name 

        :param name: name of the slot
        :return: Slot
        """
        return cls._instances[name]

    def receive(self,signal,verbose=False):
        """
        slot receives a signal

        :param signal: a Signal instance who triggered this slot to receive
        :param verbose: enables verbose mode (default False)

        """
        assert signal.name == self.name, "Signal and slot of different name have been connected, which is not allowed."
        if verbose:
            print("Slot %s received %s" % (self,signal))
        self.data[Clock().get_time()] = signal.value()

    def check_up_to_date(self):
        """
        checks if the slot is up to date, i.e. if the data has been filled with a value
        in the current time step
        """

        if Clock().get_time() in self.data:
            return True
        else:
            return False

    def value(self):
        """
        get current value
        """

        vals = self.data.values()
        if len(vals) == 0:
            warnings.warn("Empty data in Slot %s. Returning zero" % self.name)
            return 0.0

        return list(vals)[-1]

    def rlch(self):
        """
        get relative change since last update

        """
        vals = self.data.values()

        if len(vals) < 2:
            return 0

        vals = list(vals)[-2:]

        dif = (vals[-1] - vals[-2]) / vals[-2]
        return dif



class Signal:
    """
    Signal is the emitter of an information stream
    """
    _instances = {}

    def __init__(self, name, asynchronous=False):
        """
        This is a basic signal.

        :param name: name of the signal
        :param asyncronous: check_up_to_date will return True always if this is set to True
        """

        self.name = name
        self.data = {}
        self.asynchronous = False
        self.__class__._instances[self.name] = self
        self.slots = []

    def __repr__(self):
        return "[.oO SIGNAL %s]" % self.name
    @classmethod
    def retrieve(cls, name):
        """
        find a certain signal by name

        :param name: name of the signal
        :return: Signal
        """
        return cls._instances[name]

    def connect_to(self, list_of_slots):
        """
        slots connected
        
        :param list_of_slots: a list of slots to connect to this signal
        """
        self.slots = list_of_slots
        return self

    def emit(self,verbose=False):
        """
        emit the signal to all connected slots
        
        :param verbose: enables verbose mode (default False)
        """

        for slot in self.slots:
            slot.receive(self, verbose=verbose)

    def check_up_to_date(self):
        """
        checks if the signal is up to date, i.e. if the signal has been filled with a value
        in the current time step
        """
        
        if Clock().get_time() in self.data:
            return True
        else:
            return False

    def update(self, value):
        """
        insert latest value

        :param: value
        """

        self.data[Clock().get_time()] = value
        return self

    def value(self):
        """
        get current value
        """

        vals = self.data.values()
        if len(vals) == 0:
            warnings.warn("Empty data in Signal %s. Returning zero" % self.name)
            return 0.0
        return list(vals)[-1]

    def rlch(self):
        """
        get relative change since last update
        """

        vals = self.data.values()

        if len(vals) < 2:
            return 0

        vals = list(vals)[-2:]

        dif = (vals[-1] - vals[-2]) / vals[-2]
        return dif


class SignalSlot:
    """
    A connection including a signal and slot
    """

    def __init__(self, name, asynchronous=False):
        """
        
        :param name: name of the signal-slot connection
        :param asynchronous: is the signal triggered in every simulation period?
        """

        self.signal = Signal(name, asynchronous)
        self.slot = Slot(name, asynchronous)
        self.signal.slots.append(self.slot)

    def trigger(self):
        """
        Trigger of this connection. Emits the signal to the slot.
        """

        self.signal.emit()




__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

import numpy as np
from collections import defaultdict
import pandas as pd
import copy
from enum import Enum

class ICSEntry(Enum):
    """
    Enum for income sheet entry (ICSEntry). Explanation: see IncomeStatement class
    """
    REVENUES = 0
    GAINS = 1
    EXPENSES = 2
    LOSSES = 3
    NONTAX_PROFITS = 4
    NONTAX_LOSSES = 5
    INTEREST = 6
    TAXES = 7
    NOI = 8


class IncomeStatement:
    """
    Income statement data structure. Consist of the following entries:

+------------------------------------+------------------------------------------------------------------------------+
| Income Statement                   |  Example / Description                                                       |
+===========================+========+===========+==================================================================+
|  Revenues and Gains       |                    |   Money a company actually receives during a specific period     |
|                           |                    |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |   Tag  |    Value  |     interest on deposits, sales                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|  Expenditures and Losses  |                    |      wages                                                       |
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|  Gross (Before-Tax)       |                    |                                                                  |
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
| ...                       |                    |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|  Interest Payments        |                    |  interst on loans                                                |
|                           |   Tag  |    Value  |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
| ...                       |                    |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|   inter                   |                    |    interest payments                                             |
|   NOI                     |                    |    non-operating income                                          |
+---------------------------+--------------------+------------------------------------------------------------------+
|   Tax                     |                    |    taxes (e.g. on income) paid                                   |
|   paid                    |                    |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+
|                           |                    |                                                                  |
|   Non-Taxable             |                    |    untaxed profits and losses                                    |
+---------------------------+--------------------+------------------------------------------------------------------+
|   After-                  |                    |                                                                  |
|   tax income              |                    |                                                                  |
+---------------------------+--------------------+------------------------------------------------------------------+


    see also https://www.investopedia.com/terms/i/incomestatement.asp#revenues-and-gains

    """

    def __init__(self, owner,data=None,tot_dict=None,total_spendings=0.0,last=None,is_empty=True):
        """
        Constructs a new income statement.

        :param owner: Agent instance. The owner of this income statement.
        :param data: default None, can be used to pre-initialize income sheet with data. Should not be used in most cases (two-fold dict data[kind:ICSEntry][tag:str])
        :param tot_dict: default None, can be used to pre-initialize income sheet with data. Should not be used in most cases (two-fold dict tot_dict[kind:ICSEntry][tag:str]). Saves 'total' row of the sheet.
        :param total_spendings: depreciated, to not used
        """

        self.owner = owner
        self.data = data
        self.tot_dict = tot_dict
        self.last = last  # stores income statement of previous period

        if is_empty:
            self.total_spendings = 0.0 # TODO < logic behind this? Probably depreciated... # TODO safely delete
            self.restore()

    #@property
    #def taxable_income(self):
    #    """ agi + int - noi
    #     adj. gross income + interest - non-operating income
    #    """
    #    return self.gross_income - self.tot_dict[ICSEntry.INTEREST] +\
    #           self.tot_dict[ICSEntry.NOI]

    @property
    def gross_income(self):
        """agi"""
        return self.tot_dict[ICSEntry.REVENUES]\
               + self.tot_dict[ICSEntry.GAINS]\
               - self.tot_dict[ICSEntry.EXPENSES]\
               - self.tot_dict[ICSEntry.LOSSES]

    @property
    def noi(self):
        """ non-operational income"""
        return self.tot_dict[ICSEntry.NOI]

    @property
    def ebit(self):
        """earnings before interest and tax"""
        return self.gross_income\
               + self.tot_dict[ICSEntry.NONTAX_PROFITS]\
               - self.tot_dict[ICSEntry.NONTAX_LOSSES]\
               + self.noi

    @property
    def int(self):
        """interest expenditure (absolute value)"""
        return self.tot_dict[ICSEntry.INTEREST]

    @property
    def ebt(self):
        """earnings before tax (and after interest)"""
        return self.ebit + self.int

    @property
    def tax(self):
        """taxes"""
        return self.tot_dict[ICSEntry.TAXES]

    @property
    def net_income(self):
        """eranings after tax (and after interest)"""
        return self.ebt - self.tax

    @property
    def gross_spendings(self):
        """expenses + losses"""
        return self.tot_dict[ICSEntry.EXPENSES] + self.tot_dict[ICSEntry.LOSSES]

    @property
    def spendings(self):
        """expenses + losses + taxes + interest"""
        return self.tot_dict[ICSEntry.EXPENSES] + self.tot_dict[ICSEntry.LOSSES] + self.tot_dict[ICSEntry.TAXES] +\
                self.tot_dict[ICSEntry.INTEREST]

    def get_entry(self,kind,tag):
        """
        get an specific entry of the income statement

        :param kind: BalanceEntry, which entry to look up
        :param tag: str, tag of the specific row
        """
        return self.data[kind][tag]

    def new_entry(self, kind, tag, value):
        """
        Registers a new item in the income statement

        :param kind: kind of entry: Revnue, Gain, Expense, Loss or Tax
        :param tag: a name, e.g. 'interest on deposits'
        :param value: nominal value of the entry

        Example:

        .. code-block:: python

            from sfctools import Agent, ICSEntry

            a = Agent()
            ics = a.income_statement
            # ics.new_entry(kind,tag,value)
            ics.new_entry(ICSEntry.EXPENSES, "Vacation", 1000.0)
            # ...
            ics.reset()
        """

        if type(kind) == str:
            # THIS IS SLOW and should be avoided
            # introduce warning in version 0.3

            lookup = {"Revenue": ICSEntry.REVENUES,
                      "Gain": ICSEntry.GAINS,
                      "Expense": ICSEntry.EXPENSES,
                      "Loss": ICSEntry.LOSSES,
                       "Interest": ICSEntry.INTEREST,
                       "Non-Op. Income": ICSEntry.NOI,
                       "Tax": ICSEntry.TAXES,
                       "Nontax. Profit": ICSEntry.NONTAX_PROFITS,
                       "Nontax. Loss": ICSEntry.NONTAX_LOSSES
                       }

            if kind in lookup:
                kind = lookup[kind]
            else:
                error = KeyError("Passed wrong kind of income entry key. Allowed keys are %s" % str(list(lookup.values())))
                raise error
        
	  

        assert type(kind) == ICSEntry, "Passed wrong kind %s but should be %s"% (type(kind), ICSEntry)
        # ^this is indeed an error, no warning

        self.data[kind][tag] += value
        self.tot_dict[kind] += value

    def restore(self):
        """
        Restore an all-zero income statement
        """
        self.data = {ICSEntry.REVENUES: defaultdict(lambda: 0.0),
                     ICSEntry.GAINS: defaultdict(lambda: 0.0),
                     ICSEntry.EXPENSES: defaultdict(lambda: 0.0),
                     ICSEntry.LOSSES: defaultdict(lambda: 0.0),
                     ICSEntry.INTEREST: defaultdict(lambda: 0.0),
                     ICSEntry.NOI: defaultdict(lambda: 0.0),
                     ICSEntry.TAXES: defaultdict(lambda: 0.0),
                     ICSEntry.NONTAX_PROFITS: defaultdict(lambda: 0.0),
                     ICSEntry.NONTAX_LOSSES: defaultdict(lambda: 0.0)
                     }

        self.tot_dict = {
            ICSEntry.REVENUES: 0.0,
            ICSEntry.GAINS: 0.0,
            ICSEntry.EXPENSES: 0.0,
            ICSEntry.LOSSES: 0.0,
            ICSEntry.INTEREST: 0.0,
            ICSEntry.NOI:0.0,
            ICSEntry.TAXES: 0.0,
            ICSEntry.NONTAX_PROFITS: 0.0,
            ICSEntry.NONTAX_LOSSES: 0.0
        }

    def reset(self):
        """
        Reset routine. Remove data and start a new, blank income statement for the next period.
        Saves current state in "last".
        """
        self.last = IncomeStatement(owner=self.owner,data=self.data.copy(),tot_dict=self.tot_dict.copy(),total_spendings=self.total_spendings,is_empty=False) # < TODO test this properly
        self.restore()

    def to_dataframe(self):
        """
        Converts the data structure of the income sheet to human-readable pandas dataframe format.
        :return: pandas dataframe

        """
        idx = "Value"

        keys = []

        df_revenues = pd.DataFrame(self.data[ICSEntry.REVENUES], index=[idx]).T
        df_gains  =   pd.DataFrame(self.data[ICSEntry.GAINS], index=[idx]).T
        df_expenses = pd.DataFrame(self.data[ICSEntry.EXPENSES], index=[idx]).T.multiply(-1)
        df_losses =   pd.DataFrame(self.data[ICSEntry.LOSSES], index=[idx]).T.multiply(-1)
        df_total = pd.DataFrame({"Total": self.gross_income}, index=[idx]).T
        df_interest = pd.DataFrame(self.data[ICSEntry.INTEREST], index=[idx]).T
        df_noi = pd.DataFrame(self.data[ICSEntry.NOI], index=[idx]).T
        df_taxes = pd.DataFrame(self.data[ICSEntry.TAXES], index=[idx]).T.multiply(-1)
        df_nontaxable_profits = pd.DataFrame(self.data[ICSEntry.NONTAX_PROFITS], index=[idx]).T
        df_nontaxable_losses = pd.DataFrame(self.data[ICSEntry.NONTAX_LOSSES], index=[idx]).T.multiply(-1)
        df_net = pd.DataFrame({"Total": self.net_income}, index=[idx]).T

        frames = [df_revenues,df_gains,df_expenses,df_losses,
                  df_total,df_interest,df_noi,
                  df_taxes,df_nontaxable_profits,
                  df_nontaxable_losses,df_net]

        df = pd.concat(frames,keys=["Revenues","Gains","Expenses","Losses","Gross Income",
                                     "Interest","Non-Operating Income","Taxes","Nontaxable Profits",
                                     "Nontaxable Losses","Net Income"])

        df.index.names = ["Income statement of %s" %self.owner,""]
        return df

    def to_string(self):
        """
        String representation of the income sheet. This will first convert to pandas dataframe, then to string.
        :return: str
        """
        return  "\n\n" + self.to_dataframe().to_string() + "\n\n"



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod



from collections import defaultdict
import numpy as np


class WorkerRegistry:
    """
    The (rudimentary) worker registry keeps track of all agents that work at a firm (or other agent)

+----------+
|  Worker  |
+----------+
|   Agent1 |
+----------+
|   Agent2 |
+----------+
|   ...    |
+----------+
|          |
|   AgentN |
+----------+

    """

    def __init__(self, owner,wage_attr="reservation_wage"):
        """
        constructor for worker registry
        """
        self.owner = owner

        self.agent_data = []
        self.total_workforce = 0.0
        self.wage_bill = defaultdict(lambda: 0.0) # stores the wage expenditures by time

        self.wage_attr = wage_attr

    @property 
    def workers(self):
        """
        returns a list of workers
        """
        return self.agent_data

    def log_wage_bill(self,t, wb):
        """
        stores a value for wage bill wb at time t
        """
        self.wage_bill[t] = wb
        
    def pop(self):
        """
        gets the last worker 
        """
        
        return self.agent_data.pop()
        
    def get(self,i):
        """
        get worker by index 
        """
        return self.agent_data[i]
        
    def add_worker(self ,agent,verbose=False):
        """
        add a worker and corresponding wage to the registry
        """
        if verbose:
            print(self.owner,"add worker",agent)
        agent.employer = self.owner
        self.agent_data.append(agent)
        self.total_workforce += 1

    def remove_worker(self, agent):
        """
        remove a worker from registry (has been fired, retired, ...)
        """

        if hasattr(self.owner,"labor_market"):
            self.owner.labor_market.unlink_agents(agent, agent.employer)

        agent.is_employed = 0.0
        agent.employer = None

        self.agent_data.remove(agent)
        self.total_workforce -= 1

    def get_avg_costs(self):
        """
        computes the cost of labor, i.e. the average value of all wages
        """
        # each worker works 1 unit
        vals = [getattr(worker,self.wage_attr) * 1.0 for worker in self.agent_data]

        if len(vals) > 0:

            return np.mean(vals)
        else:
            return 0

    def fire_random(self, number,verbose=False):
        """
        fire a certain number of workers at random
        """

        if verbose:
            print("fire",self.agent_data)

        for i in range(min(number, len(self.agent_data))):
            random_worker = np.random.choice(self.agent_data)
            self.remove_worker(random_worker)

    def fire_all(self):
        """
        remove all workers from the registry
        """
        for worker in self.agent_data:
            self.remove_worker(worker)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


from .world import World

from ..datastructs.balance import BalanceSheet, BalanceEntry
from ..datastructs.income_statement import IncomeStatement
from ..datastructs.cash_flow_statement import CashFlowStatement
from ..datastructs.signalslot import Signal,Slot
from .flow_matrix import FlowMatrix


import numpy as np
import warnings
from collections import defaultdict
from functools import wraps


class Agent:
    """
    This is the base class for any agent. It takes care of the bookkeeping system
    and other low-level operations of the agent.

    """

    def __init__(self, alias=None, verbose=False, signals:list = None, slots:list = None ):
        """
        Instantiation of agents is done here.
        Note: Each agent will be given a name and an alias name. The name will be set automatically and
        the alias name will be a (possibly non-unique) name, for example 'name' could be 'Firm_0001' and
        alias_name could be 'MyEnergyProducer'.

        :param alias: (optional) alias or None(default)
        :param verbose: boolean switch to enable console output. Default is False
        :param signals: list of str or None(default), signal names connected to this node
        :param slots: list of str or None(default), slot names connected to this node.
        """

        self.name = "New Agent" # set dummy name for registering agent

        world = World()  # get singleton world
        world.register_agent(self) # register agent at world registry

        # set name
        self.name = str(self.__class__.__name__) + "__%05i" % world.agent_registry.get_count(self.__class__.__name__)

        if alias is None: # set the alias if desired
            self.alias = self.name
        else:
            self.alias = alias

        self._balance_sheet = BalanceSheet(self) # agent gets a blank balance sheet
        self.income_statement = IncomeStatement(self) # blank income statement
        self.cash_flow_statement = CashFlowStatement(self)  # blank cash flow statement

        # triggers for tree struct to enable event-based approach
        self.trigger_dict = {} # <- this is meant for automation and therefore cannot be set by the user. TODO re-think implementation

        # bankruptcy flag
        self.bankrupt = False

        # print notification in verbose mode
        if verbose:
            if self.name != self.alias:
                print("New Agent %s" % self.name, "alias", self.alias)
            else:
                print("New Agent %s" % self.name)

        # triggers for signals and slots
        if signals is not None or slots is not None:
            self.signals = {}  # agent only has attribute "signals" if given
            self.slots = {}    # agent only has attribute "slots" if given

            # fill signals and slots dictionaries...
            assert isinstance(signals,list) or signals is None, "signal must be a list of str or None"
            if signals is not None:
                for signal_name in signals:
                    self.signals[signal_name] = Signal.retrieve(signal_name)

            assert isinstance(slots, list) or slots is None, "signal must be a list of str or None"
            if slots is not None:
                for slot_name in slots:
                    self.slots[slot_name] = Slot.retrieve(slot_name)

    def trigger(self, event_key,*args,**kwargs):
        """
        trigger a certain default method of this agent
        :param event_key str: a key for the event
        """
        method = self.trigger_dict[event_key]
        method()

    @property
    def balance_sheet(self):
        """
        This will return the (private) BalanceSheet object of the agent.
        """
        return self._balance_sheet

    def __repr__(self):
            s = ""
            if hasattr(self, "alias") and self.name != self.alias:
                s = " alias " + self.alias

            return "<Agent: " + self.name + s + ">"

    def __str__(self):
        if hasattr(self, "alias") and self.alias is not None and self.name != self.alias:
            return self.name + "(" + self.alias + ")" # + " [ %i ]"%id(self)
        else:
            return self.name # + " [ %i ]"%id(self)

    def __lt__(self, other): # required for pandas
        return self.name < other.name

    @property
    def leverage(self) -> float:
        """
        leverage of own balance sheet
        """
        return self.balance_sheet.leverage

    @property
    def net_worth(self) -> float:
        """
        net worth at own balance sheet
        """
        return self.balance_sheet.net_worth

    @property
    def total_assets(self):
        """
        sum of assets column of balance sheet
        """
        return self.balance_sheet.total_assets

    @property
    def total_liabilities(self):
        """
        sum of liabilities column of balance sheet
        """
        return  self.balance_sheet.total_liabilities

    @property
    def cash_balance(self) -> float:
        """
        'Shortcut' property for balance of 'Cash' stored in 'Assets' of the agent
        """
        return self.balance_sheet.get_balance("Cash", BalanceEntry.ASSETS)

    def file_bankruptcy(self, event=None):
        """ file agent for bankrupcy

        :param event: preferrably str, can be None for manual trigger but should be something like 'negative cash' or 'negative equity' (default).

        NOTE This should be called by a sub-module rather than manually by the user
        """
        self.bankrupt = True
        raise RuntimeError("%s went bankrupt (reason: %s) ! %s" %(self,event,self.balance_sheet.to_string()))

    def get_balance(self, key,kind) -> float:
        """
        Gets the nominal balance of a certain entry in the balance sheet.
        :param key: str, which item is requested (e.g. 'Cash', 'Apples',...)
        :param kind: BalanceEntry
        """
        # get balance sheet asset balance
        return self.balance_sheet.get_balance(key,kind=kind)



class SingletonAgent(Agent):
    """
    SingletonAgent is an agent which is singleton, meaning that only one instance creation of this Agent will be allowed.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(self,alias=None, verbose=False):
        super().__init__(alias=alias, verbose=verbose)


def block_on_bankrupt(method):
    """
    This is a decorator you can use when building an agent.
    It will block this action whenever the agent is bankrupt.

    Usage:

    .. code-block:: python

        from sfctools import block_on_bankrupt

        class MyAgent(Agent):
            ...

        @block_on_bankrupt
        def my_fun(self,...):
            ...

    """
    @wraps(method)
    def _impl(self,*args,**kwargs):
        if not self.bankrupt:
            method_output = method(self,*args,**kwargs)
            return method_output
        else:
            warnings.warn("%s: tried to call forbidden method in bankruptcy: %s"%(self,method)) # <- TODO test this
            return None
    return _impl



__version__ = "0.4"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

"""
Small plotting library for bar plots and line plots. Usage is optional / voluntary...
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pylab import rcParams
import matplotlib.patches as mpatches
import matplotlib as mpl
from functools import partial


def add_value_labels(ax, spacing=.9, fmt="{0:f}", stacked=False, legend="best"):
    """Add labels to the end of each bar in a bar chart.

    Arguments:
        ax (matplotlib.axes.Axes): The matplotlib object containing the axes
            of the plot to annotate.
        spacing (int): The distance between the labels and the bars.
    """

    my_figure = ax.get_figure()
    L = abs(ax.get_ylim()[1] - ax.get_ylim()[0])
    min_height = L*0.015
    small_height = L*0.05
    # print("min_heigth", min_height)

    # For each bar: Place a label
    for rect in ax.patches:
        # Get X and Y placement of label from rect.
        outside = False

        if abs(rect.get_height()) <= min_height:
            continue

        elif abs(rect.get_height()) <= small_height:
            outside = True
            y_value = rect.get_y() + rect.get_height() * 0.1
            x_value = rect.get_x() + rect.get_width() * 1.05

        # print("rect height", rect.get_height())
        else:
            y_value = rect.get_y() + rect.get_height() * 0.41
            x_value = rect.get_x() + rect.get_width() * 0.50

        # print("yvalue",x_value,y_value)
        # Number of points between bar and label. Change to your liking.

        space = spacing
        # Vertical alignment for positive values
        va = 'bottom'

        # If value of bar is negative: Place label below bar
        if y_value < 0:
            # Invert space to place label below
            space *= -1
            # Vertically align label at top
            va = 'top'

        # Use Y value as label and format number with one decimal place
        # label = "{:.1f}".format(y_value)
        hgt = rect.get_height()
        if hgt != 0.0:
            label = fmt.format(hgt)
        else:
            label = ""

        # Create annotation
        if outside:
            rot = 33
            ha = "left"
        else:
            rot = 0
            ha = "center"

        t = ax.annotate(
            label,  # Use `label` as label
            (x_value, y_value),  # Place label at end of the bar
            xytext=(0, space),  # Vertically shift label by `space`
            textcoords="offset points",  # Interpret `xytext` as offset in points
            ha=ha,  # Horizontally center label
            va=va)  # Vertically align label differently for

        if not outside:
            t.set_bbox(dict(facecolor='white', alpha=0.175, edgecolor='white',linewidth=0.0))
        # positive and negative values.


def matplotlib_barplot(data, xlabel, ylabel, title, color="indigo", size=(6, 5), tight_layout=True,
                       fmt="{0:f}", stacked=False, show_labels=True, legend="best", show=True,yerr=None,ax=None):
    """
    Creates a bar plot in matplotlib. The 'macro' plotting style will be used if it is set up

    :param xlabel: x axis  label
    :param ylabel: y axis  label
    :param title: plot title
    :param color: plot color
    :param size: figure size
    :param tight_layout: re-format plot
    :param fmt: number format, {0:d} or {0:f} for example
    :param stacked: stack column? or show side-by side (boolean)
    :paarm show_labels: show the column labels at bottom? (boolean)
    :param legend: location of legend, or 'off'
    :param show: if False, only the figure is returned. Else plot is shown
    :param ax: axis to plot on
    :param yerr: data for error bar (if any)
    """

    # if legend is 'off', no legend is shown

    plt.rcParams['axes.ymargin'] = .4
    rcParams['figure.figsize'] = size

    try:
        plt.style.use("macro") # if 'macro' plotting style is installed -> use it!
    except:
        pass

    plt.grid(alpha=0.0)

    if not isinstance(data, pd.DataFrame):
        if ylabel is None:
            ylabel = "Data"
        data = pd.DataFrame({ylabel: data})
        data.set_index(ylabel)

    # plt.figure(figsize=size)
    if ax is None:
        ax = plt.gca()
    data.plot.bar(color=color, stacked=stacked,ax=ax,yerr=yerr,capsize=6,ecolor="black")
    plt.gca().set_title(title)
    # plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=True)
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)

    if show_labels:
        add_value_labels(plt.gca(), fmt=fmt)

    if legend != "off":
        plt.legend(loc=legend)
    else:
        old_legend = plt.gca().get_legend()
        if old_legend is not None:
            old_legend.remove()
    if tight_layout:
        plt.tight_layout()

    # fix reverse labeling
    # https://stackoverflow.com/questions/46908085/bar-plot-does-not-respect-order-of-the-legend-text-in-matplotlib
    #
    ax = plt.gca()
    h, l = ax.get_legend_handles_labels()
    if legend != "off":
        ax.legend(h[::-1], l[::-1],loc=legend,framealpha=0.0)

    if show:
        plt.show()
    else:
        return plt.gcf()


def matplotlib_lineplot(data, xlabel=None, ylabel=None, title="", xlim=None, ylim=None, color="indigo", legend="best",marker=None,show=False,ax=None):
    """
    creates a line plot in matplotlib. The 'macro' plotting style will be used if it is set up

    :param xlabel: x axis  label
    :param ylabel: y axis  label
    :param title: plot title
    :param color: plot color
    :param xlim, ylim: tuples of plot limits
    :param legend: location of legend, or 'off', or 'outside'
    :param show: if False, only the figure is returned. Else plot is shown
    :param ax: axis to plot on

    when overlapping barplot and lineplot, make sure you add an additional axis, see
    https://stackoverflow.com/questions/42948576/pandas-plot-does-not-overlay

    """
    try:
        plt.style.use("macro")
    except:
        pass


    if not isinstance(data, pd.DataFrame):
        if ylabel is None:
            label = "Data"
        data = pd.DataFrame({ylabel: data})
        data.set_index(ylabel)

    if ax is None:
        ax = plt.gca()

    for col in data.columns:
        ax.plot(data[col],color=color,marker=marker,label=col)

    plt.xlim(xlim)
    plt.ylim(ylim)
    plt.title(title)

    if legend != "off":
        plt.legend(loc=legend)
    elif legend == "outside":
        plt.legend(loc=(1.02,0))

    elif legen == "off":
        old_legend = plt.gca().get_legend()
        if old_legend is not None:
            old_legend.remove()

    if ylabel is not None:
        plt.ylabel(ylabel)

    if xlabel is not None:
        plt.gca().set_xlabel(xlabel)

    if ylabel is not None:
        plt.gca().set_ylabel(ylabel)
    plt.rcParams['axes.ymargin'] = .4

    if show:
        plt.show()
    else:
        return plt.gcf()





class Point:
    """
    a rudimentary point in 2d space
    """

    def __init__(self, x, y):
        self._x = x
        self._y = y
        self.offset_y = 0 # offset for bands

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y + self.offset_y

class Label:
    """
    a rudimentary node label
    """

    names = []

    def __init__(self,x,y,text):
        self.x = x
        self.y = y
        self.text = text
        if text in self.__class__.names:
            raise RuntimeError("Name already given")
        else:
            self.__class__.names.append(self.text)

    def shift_up(self):
        # shift the label upwards a bit
        self.y += 1.8
        self.x -= 0.2 # re-center


def casteljau(t,b0,b1,b2,b3):
    """
    casteljau bezier spline for 4 points.

    :param b0-b3: the point of the bezier curve as [x,y] array
    :param t: parameter for position on the curve, float
    """
    b = np.multiply((-b0 + 3*b1 - 3*b2 + b3),np.power(t,3))  + np.multiply((3*b0 - 6*b1 + 3*b2),np.power(t,2)) + np.multiply((-3*b0 + 3*b1),t) + b0
    return b


def bezier(pstart : Point, pend: Point ,npoints = 100, curv=None):
    """
    compute the path of a 1d bezier curve

    :param pstart: starting point of the bezier curve
    :param pend: end point of the bezier curve
    :param npoints: number of discrete points to retrun
    :param curv: optionally you can give a curvature parameter here (float) default None

    :return: x,y coordinate tuple as ndarray of shape (npoints,)
    """

    # de-casteljau b-spline
    tarr = np.linspace(0,1,npoints) # parameter of the curve position
    # define 4 points: b0, b1, b2, b3

    if pstart.x < pend.x: # pstart left of pend?
        pleft = pstart
        pright = pend
    else:
        pleft = pend
        pright = pstart

    if curv is None:
        curv = .27*(pright.x-pleft.x)

    b0 = np.array([pleft.x,pleft.y])
    b1 = np.array([pleft.x+curv,pleft.y])
    b2 = np.array([pright.x-curv,pright.y])
    b3 = np.array([pright.x,pright.y])

    spline = partial(casteljau,b0=b0,b1=b1,b2=b2,b3=b3)

    b = np.array([spline(t) for t in tarr])

    return b[:,0],b[:,1]


def plot_band(pstart: Point, pend: Point, npoints=100,curv=None, weight=2):
    """
    computes a band of two bezier curves

    :param pstart: starting point of the bezier curve
    :param pend: end point of the bezier curve
    :param npoints: number of discrete points to retrun
    :param curv: optionally you can give a curvature parameter here (float) default None
    :param weight: width of the band

    :return: x,y1,y2 coordinate 3-tuple of the band
    """

    w = weight # width of the band
    d = 0 # horizontal shift of the bands realtive to each other
    e = 0 # distance of  '<' and '>' arc

    qstart1 = Point(pstart.x-d,pstart.y)
    qstart2 = Point(pstart.x-d,pstart.y)

    qend  = Point(pend.x-d,pend.y)
    qend1  = Point(pend.x-d,pend.y)
    qend2  = Point(pend.x-d,pend.y)

    pstart1 = Point(pstart.x-d+.2*e,pstart.y)
    pend1 = Point(pend.x-d-e,pend.y)

    pstart2 = Point(pstart.x+d+.2*e,pstart.y+w)
    pend2 = Point(pend.x+d-e,pend.y+w)

    bezier_start1 = bezier(qstart1,pstart1)
    bezier_start2 = bezier(qstart2,pstart2)

    bezier_end1 = bezier(pend1,pend1)
    bezier_end2 = bezier(pend2,pend2)

    bezier1 = bezier(pstart1,pend1)
    bezier2 = bezier(pstart2,pend2)

    #x = np.array(list(bezier_start1[0])+list(bezier1[0])+list(bezier_end1[0]))
    #y1 = np.array(list(bezier_start1[1])+list(bezier1[1])+list(bezier_end1[1]))
    #y2 = np.array(list(bezier_start2[1])+list(bezier2[1])+list(bezier_end2[1]))

    x = np.array(list(bezier1[0]))  # + list(bezier_end1[0]))
    y1 = np.array(list(bezier1[1])) #  + list(bezier_end1[1]))
    y2 = np.array(list(bezier2[1])) # + list(bezier_end2[1]))


    return x, y1, y2


def draw_band(fig,pstart:Point, pend:Point, color="blue",alpha=.9,weight=2,norm=1.0,show_label=True,label=None,dv=0.6,rotation=0):
    """
    draw a band graphically

    """
    x,y1,y2 = band(pstart,pend,weight=weight)
    plt.fill_between(x,y1,y2,color=color,figure=fig,alpha=alpha)
    
    my_bez_x, my_bez_y = bezier(Point(pstart.x, pstart.y), Point(pend.x, pend.y))
    
    plt.plot(my_bez_x, my_bez_y, color="black",alpha=.7) # "black")

    if show_label and label is not None:
        # draw bezier
        
        # dv = 0.6
        bx = my_bez_x[int(dv*len(my_bez_x))]
        by = my_bez_y[int(dv*len(my_bez_y))]

        
        #bx2,by2 = casteljau(t=0.55,b0=np.array([pstart.x,pstart.y]),b1=b1,b2=b2,b3=np.array([pend.x,pend.y]))
        #bx3,by3 = casteljau(t=0.66,b0=np.array([pstart.x,pstart.y]),b1=b1,b2=b2,b3=np.array([pend.x,pend.y]))
        #rot = (90-.5*np.arctan2((bx3-bx2),(by3-by2))*180.0/np.pi)%360.0 # compute slope at point
        #rot = 0

        bbox = {'pad': 2, 'facecolor': color,'edgecolor':color,'linewidth':.0}
        props = {'ha': 'center', 'va': 'center', 'bbox': bbox}

        if label is not None:
            plt.gca().text(bx,by - .0, label, props, rotation=rotation,color="white", alpha=alpha)
        else:
            plt.gca().text(bx,by , "%.2f" %(weight/norm), props, rotation=rotation,color="black",alpha=alpha)

        
        # plt.annotate(str(weight),(bx,by),color="gray",alpha=alpha)

    pend.offset_y += 1.5*weight
    pstart.offset_y += 1.5*weight

def plot_sankey(data,title="",show_plot=True,
                show_values=True,
                colors=None,
                label_rot=65,
                fontsize=10,
                separation=0.8,
                norm_factor=1.5,
                min_width=0.1,
                dy=10,
                dx=55,
                dx_space=8.5,
                alpha=0.5,
                filling_fraction = 0.85):
    """
    plots a sankey diagram from data

    :param data: list of pandas dataframes, of the following format. length at least two.
    :param title: title of the plot
    :param show_plot: boolean switch to show plot window (default True). If False, figure is returned
    :param show_values: boolean switch to print numerical values of data as text label (default True)
    :param colors: (optional) a list of colors (if None, default colors are chosen)
    :param label_rot: rotation of the labels (default 65 degrees)
    :param fontsize: size of label font
    :param separation: separation point of the label along the bands, between 0 and 1,
    :param norm_factor: normalization factor for width 
    :param min_width: minimum width of bands
    :param dy: vertical distance of bands
    :param dx: horizontal distance of bands
    :param dx_space: space between layers
    :param filling_fraction: fraction by which the bands align (scaling with dy)

    :return fig: None or matplotlib figure

    Example: 

    +----------+----------+----------+-------------+
    |  from    |   to     |  value   | color_id    |
    +----------+----------+----------+-------------+
    |    A     |  C       | 1.0      |    0        |
    +----------+----------+----------+-------------+
    |    A     |  D       | 2.0      |    1        |
    +----------+----------+----------+-------------+
    |    B     |  D       | 3.0      |    2        |
    +----------+----------+----------+-------------+

    +----------+----------+----------+-------------+
    |  from    |   to     |  value   | color_id    |
    +----------+----------+----------+-------------+
    |    C     |  E       | 5.0      |    1        |
    +----------+----------+----------+-------------+
    |    C     |  F       | 6.0      |    2        |
    +----------+----------+----------+-------------+
    |    D     |  G       | 8.0      |    3        |
    +----------+----------+----------+-------------+

    """

    Label.names = []

    if not isinstance(data,list):
        data = [data] # convert to list

    if colors is None:
        colors = [
            "lightgray",
            "skyblue",
            "wheat",
            "lightgreen",
            "lavender",
        ]
    
    # 1. normalize the values to reasonable width of bands

    # norm_factor = 1.5 #normalization factor for width

    maxval = 0.0
    maxrows = 2

    for layer in data:

        maxrows = max(maxrows,len(list(layer.iterrows())))
        for idx,row in layer.iterrows():
            val = float(row["value"])

            maxval = max(val,maxval)

    # norm_factor = 1.5
    if maxval > 0.0:
        norm_factor = 1.0 /maxval

    fig = plt.figure(figsize=(2.4*len(data),1.05*maxrows)) # 1.8*maxrows

    # 2. plot the bezier patches

    i = 10  #
    j0 = 6 # < starting position (i,j)

    # dy = 10 # 8  # vertical distance of bands
    # dx = 55 #   # horizontal distance of layers

    # dx_space = 2.5 # white space between layers

    # define some points for the unique nodes
    my_points = {} # keys will be names, value will be Points
    my_labels = {} # keys will be names, value will be Labels
    my_values = {} # labels for values

    last_j = j0

    # iterate through the dataframes
    # heights = defaultdict(lambda: 0 )

    
    class Band:
    

            instances = [] 
            def __init__(self,n_from,n_to,w,color="steelblue"):
                self.n_from = n_from 
                self.n_to = n_to 
                self.shift_y =0.0 
                self.width =  2* w
                self.osy1 = 0.0
                self.osy2 = 0.0
                self.color = color
                self.__class__.instances.append(self) 

            def plot(self):
                fromnode = self.n_from 
                tonode = self.n_to
                w = self.width
                osy1 = self.osy1
                osy2 = self.osy2
                
                #vals_lower = [fromnode.y+osy1,tonode.y+osy2]
                #vals_upper = [fromnode.y+osy1 + w,tonode.y+osy2 + w]

                point_lower1 = Point(fromnode.x,fromnode.y+osy1)
                point_lower2  = Point(tonode.x, tonode.y + osy2)
                # point_lower2 = Point(fromnode.x,fromnode.y+osy1+w)

                # make the curves more smooth 
                x, vals_lower, vals_upper = plot_band(point_lower1, point_lower2,weight=w)
                plt.fill_between(x, vals_lower, vals_upper,color=self.color)
                
                
            def maxy(self):
                w = 2 * self.width 
                return max([self.n_from.y+ self.osy1 + w, self.n_to.y+ self.osy2 + w])
            

   

    from collections import defaultdict
    bands = defaultdict(lambda: [])

    
    for indx, layer in enumerate(data):

        j =  j0 

        k = 0
        for name in layer["from"].unique():

            if name not in my_points:
                new_point = Point(i+dx_space,j)
                my_points[name] = new_point

                j += dy

            if name not in my_labels:
                my_labels[name] = Label(new_point.x-5.2,new_point.y-.2,name) # plt.annotate(name,(new_point.x-1.8, new_point.y-.1))
            else:

                my_labels[name].shift_up()

            k+=1

        j = j0

        k = 0
        for name in layer["to"].unique():

            if name not in my_points:
                new_point = Point(i+dx,j)
                my_points[name] = new_point #Point(i+dx,j)

                j += dy 

            if name not in my_labels:
                my_labels[name] = Label(new_point.x+.5*dx_space,new_point.y-.2,name) # plt.annotate(name,(new_point.x+.5*dx_space,new_point.y-.1))
            else:
                my_labels[name].shift_up()
            
            k += 1

        last_j = j - dy

        for idx, row in layer.iterrows():

            fromnode = my_points[row["from"]]
            tonode = my_points[row["to"]]
            color_idx = row["color_id"]

            my_band = Band(fromnode, tonode, w = float(row["value"])*norm_factor * dy * (1-filling_fraction), color = colors[color_idx%len(colors)])
            
            bands[fromnode].append(my_band)
            bands[tonode].append(my_band)

        i += dx

        #for name,point in my_points.items():
        #    point.offset_y = 0

        # iterate through the dataframe rows


        for point, b in bands.items():
             
            offset1 = 0.0 
            offset2 = 0.0 

            for bi in b:

                if point == bi.n_from:
                    bi.osy1 = offset1
                    offset1 += bi.width 

                elif point == bi.n_to:
                    bi.osy2 = offset2
                    offset2 += bi.width 
          

        # plot all bands 
        for band in Band.instances:
            band.plot()

        """
        for idx, row in layer.iterrows():

            color_idx = int(row["color_id"]) # index for color map

            fromnode = my_points[row["from"]]
            tonode = my_points[row["to"]]
            weight = float(row["value"])*norm_factor # 8.5e-2 #  max(min_width,float(row["value"])*norm_factor) # <- make bars thicker

            draw_band(fig,fromnode,tonode,colors[color_idx%len(colors)],weight=weight,norm=norm_factor,show_label=show_values,rotation=0,dv=separation,label=str(row["value"]))

            # update weight for next layer
        """


    for label in my_labels.values(): # draw the labels here...
        if len(label.text) <= 2:
            plt.annotate(label.text.replace("$[source]","").replace("$[sink]",""),(label.x,label.y))
        else: # rotate label for better visibility
            plt.gca().text(label.x,label.y, label.text.replace("$[source]","").replace("$[sink]",""), rotation=label_rot,fontsize=fontsize)

    # 3. return the figure or show window

    plt.axis("off")

    max_x = 0
    max_y = 0

    for point in my_points.values():
        max_x = max(max_x,point.x)
        max_y = max(max_y,point.y)

    for label in my_labels.values():
        max_x = max(max_x,label.x)
        max_y = max(max_y,label.y)

    maxy = max([b.maxy() for b in Band.instances])
    plt.xlim([3,max_x+5])
    plt.ylim([3,maxy+3])

    #plt.autoscale()

    plt.tight_layout()

    plt.title(title)

    plt.tight_layout()

    if show_plot:
        plt.show()

    return fig



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


from abc import ABCMeta,abstractmethod
from collections import defaultdict
import networkx as nx
import numpy as np
from ..core.singleton import Singleton
import warnings
import matplotlib.pyplot as plt

class MarketMatching(metaclass=ABCMeta):
    """
    Meta class for market matching. This is a basic framework for a list of demanders and suppliers who are matched decentrally and
    registered centrally. This can be thought of as a graph (nodes and edges) of agents.
    """

    def __init__(self,demand_list=None,supply_list=None):
        """
        constructor for market matching

        :param demand_list: list of demand agents or None (default, creates empty list)
        :param supply_list: list of supply agents or None (default)
        """

        super().__init__()

        # matching map between demanders and suppliers
        self.graph = nx.MultiDiGraph()

        if demand_list is None:
            demand_list = []
        if supply_list is None:
            supply_list = []

        self._demand_list = demand_list
        self._supply_list = supply_list


    @property
    def supply_list(self):
        """
        Get list of supply agents. Filters suppliers who are not bankrupt

        :return: list of agents
        """

        my_list = []

        for agent in self._supply_list:
            if not agent.bankrupt:
                my_list.append(agent)

        return my_list

    @property
    def demand_list(self):
        """
        Get list of demand agents. Filters demanders who are not bankrupt

        :return: list of agents
        """

        my_list = []
        for agent in self._demand_list:
            if not agent.bankrupt:
                my_list.append(agent)

        return my_list

    def add_demander(self,agent):
        """
        Add demand agent as node in the graph

        :param agent: instance to add on demand side
        """
        self.graph.add_node(agent)
        self._demand_list.append(agent)

    def add_supplier(self, agent):
        """
        Add supply agent as node in the graph

        :param agent: instance to add on demand side
        """
        self.graph.add_node(agent)
        self._supply_list.append(agent)

    @abstractmethod
    def rematch(self):
        """
        rematch the supply and demand agents, i.e.
        update the matching map. Has to be overridden as it is abstract.
        """
        pass

    def clear(self):
        """
        Removes all edges from the current graph.
        """

        e = list(self.graph.edges())
        self.graph.remove_edges_from(e)

    def clear_all(self):
        """
        Do a complete reset (remove edges and nodes, reset to blank greenfield market.
        """
        self.graph = nx.MultiDiGraph()
        self._demand_list = []
        self._supply_list = []


    def get_supply_data(self,agent):
        """
        get all suppliers  + info for agent

        :param agent: reference agent
        :return: dict containing data
        """
        in_edges = self.graph.in_edges(agent)
        supply_data = {}
        for edge in in_edges:
            supply_data[edge[0]] = next(iter(self.graph.get_edge_data(*edge).keys()))
        return supply_data

    def get_demand_data(self,agent):
        """
        get all demanders + info for agent

        :param agent: reference agent
        :return: dict containing data
        """
        in_edges = self.graph.out_edges(agent)
        supply_data = {}
        for edge in in_edges:
            supply_data[edge[1]] = next(iter(self.graph.get_edge_data(*edge).keys()))
        return supply_data

    def get_suppliers_of(self,agent) -> list:
        """
        get list of agents supplying to agent

        :param agent: reference agent
        :return: list of agents
        """
        if agent in self.graph.pred:
            return list(self.graph.pred[agent].keys())
        return []

    def get_demanders_from(self,agent) -> list:
        """
        get list of agents demanding from agent

        :param agent: reference agent
        :return: list of agents
        """
        if agent in self.graph:
            return list(self.graph[agent].keys())
        return []

    def link_agents(self,supply_agent,demand_agent,val):
        """
        link a connection between demand and supply with weight val

        :param supply_agent: agent instance, supply
        :param demand_agent: agent instance, demand
        :param val: a value that is stored in the data dict for this link
        """
        self.graph.add_edge(supply_agent,demand_agent,weight=val)

    def unlink_agents(self,supply_agent,demand_agent):
        """
        unlink a connection

        :param supply_agent: agent instance, supply
        :param demand_agent: agent instance, demand
        """
        try:
            self.graph.remove_edge(supply_agent,demand_agent)
        except Exception as e:
            warnings.warn(str(e))

    def remove_supplier(self,agent):
        """
        remove a supplier
        """
        self._supply_list.remove(agent)

    def remove_demander(self,agent):
        """
        remove a demander
        """
        self._demand_list.remove(agent)


    def get_value(self,agent_from,agent_to):
        try:
            return self.graph[agent_from][agent_to]["weight"]
        except:
            return None


    def plot(self,fname=None,show_figure=True):
        """
        plot the current matching situation as a network graph
        :param fname: filename (optional) if filename is given, no plot will be shown and a png will be saved to a file.
        :param show_figure: default True, show the figure in a plot window
        """
        import matplotlib.pyplot as plt

        nodelist = self.graph.nodes()

        colors = []
        for node in nodelist:

            if node in self._demand_list:
                colors.append("blue")
            elif node in self._supply_list:
                colors.append("red")

            else:
                colors.append("gray")
                warnings.warn("There are unknown agents in the market graph. Please add them to the demand or supply list (add_supplier or add_demander).")

        if show_figure:
            plt.figure()

        nx.draw(self.graph,node_size=20,width=.2,node_color=colors)

        if fname is None:

            if show_figure:
                plt.show()
        else:
            plt.savefig(fname)
            plt.close()


    def plot_weighted(self,my_layout=None,arrows=False,fname=None):
        """
        plot the current matching situation as a network graph, with edge thickness matching the weights
        Suppliers re plotted in red, demanders in blue.

        :param my_layout: networkx graph layout
        :param arrows: show the arrows
        :param fname: if None, plot is shown in window. Else, file is stored to fname
        """
        widths = nx.get_edge_attributes(self.graph, 'weight')
        m = max(widths.values()) # max linewidth is 3.5

        for k,v in widths.items():
            widths[k] = widths[k]/m * 3.5
        
        nodelist = self.graph.nodes()

        colors = []
        for node in nodelist:
            # print("node",node)
            if node in self._demand_list:
                colors.append("blue")

            elif node in self._supply_list:
                colors.append("red")

            else:
                colors.append("gray")
                warnings.warn("There are unknown agents in the market graph. Please add them to the demand or supply list (add_supplier or add_demander).")

        plt.figure(figsize=(5,5))
        if my_layout is None:
            my_layout = nx.drawing.layout.spring_layout
        pos = my_layout(self.graph) # ,weights="weight")
        nx.draw_networkx_nodes(self.graph,pos,
                       nodelist=nodelist,
                       node_size=20,
                       node_color=colors,
                       alpha=1.0)
        nx.draw_networkx_edges(self.graph,pos,
                       edgelist = widths.keys(),
                       width=list(widths.values()),
                       edge_color='black',
                       alpha=1.0,
                       arrows=arrows)
        plt.box(False)

        if fname is None:
            plt.show()
        else:
            plt.savefig(fname)
            plt.close()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'test' # options are: dev, test, prod


from collections import defaultdict
import pandas as pd

class RegistryEntry:
    """
    A helper data structure for MarketRegistry
    """

    def __init__(self):
        self.data = {"total": 0.0, "fulfilled": 0.0}

    def __str__(self):
        return str((self["total"],self["fulfilled"],self["remaining"]))

    def __getitem__(self, item):
        if item == "total" or item == "fulfilled":

            if callable(self.data[item]):
                call = self.data[item]
                return call()
            else:
                return self.data[item]

        if item == "remaining":
            return self["total"] - self["fulfilled"]

    def __setitem__(self, key, value):
        if key== "total" or key=="fulfilled":
            self.data[key] = value
        else:   
            raise KeyError("Cannot set key %s",key)


class MarketRegistry:
    """
    Data structure to store remaining and fulfilled demand of an agent at a market


    DEMAND:

    +---------+--------------------+-------------------+------------------------+
    |         |       total        |    fulfilled      |    remaining           |
    +---------+--------------------+-------------------+------------------------+
    | good1   |                    |                   |    =total-fulfilled    |
    +---------+--------------------+-------------------+------------------------+
    | good2   |                    |                   |                        |
    +---------+--------------------+-------------------+------------------------+
    | good3   |                    |                   |                        |
    +---------+--------------------+-------------------+------------------------+
    

    SUPPLY:

    +---------+--------------------+-------------------+------------------------+
    |         |       total        |    fulfilled      |    remaining           |
    +---------+--------------------+-------------------+------------------------+
    | good1   |                    |                   |    =total-fulfilled    |
    +---------+--------------------+-------------------+------------------------+
    | good2   |                    |                   |                        |
    +---------+--------------------+-------------------+------------------------+
    | good3   |                    |                   |                        |
    +---------+--------------------+-------------------+------------------------+

    """
    
    def __init__(self,owner):
        """
        :param owner: agent instance who owns this data structure
        """
        self.data = {"demand":defaultdict(lambda: RegistryEntry()),
                     "supply":defaultdict(lambda: RegistryEntry())}

        self.owner = owner

    def to_string(self):
        """
        converts data structure to human-readable string format
        """
        df = pd.DataFrame.from_dict(self.data)
        df.index.name = str(self.owner)   # .set_caption(self.owner)
        return "\n\n" +  df.to_string()

    def __getitem__(self, item):
        return self.data[item]

    def set_target_update(self, kind, good, new_value):
        """
        set up the toftal demand/supply for a certain good
        
        :param kind: 'demand' or 'supply'
        :param good: name of good
        :param q: new entry value
        """
        self.data[kind][good]["total"] = new_value

    def set_current_update(self, kind, good, q):
        """"
        update fulfillment of demand or supply

        :param kind: 'demand' or 'supply'
        :param good: name of good
        :param q: new entry value
        """
        self.data[kind][good]["fulfilled"] = q

    def reset(self,kind,good):
        """
        reset all entries to zero
        
        :param kind: 'demand' or 'supply'
        :param good: name of good
        """
        self.data[kind][good]["total"] = 0
        self.data[kind][good]["fulfilled"] = 0





from sfctools.automation.runner import ModelRunner
from sfctools import Settings, World, Agent
import numpy as np
np.random.seed(1235)
import pandas as pd
import os 

"""
This tests the automation routines for monte-carlo batches
"""

from sfctools.examples.example_wrapper import Example


def run(base_path):
    # run this in the upper directory via pytest tests/, otherwise this will fail


    # run this in the upper directory via pytest tests/, otherwise this will fail
    settings_path = os.path.join(base_path, "testsettings.yml")
    results_path = os.path.join(base_path, "results/")

    # run this in the upper directory via pytest tests/, otherwise this will fail

    class MyAgent(Agent):
        def __init__(self):
            super().__init__()
            self.a_param = Settings()["param_a"]

        def modify_a(self):
            self.a_param *= np.random.rand()

    def builder():
        # placeholder for agent builder, 
        my_agents = [MyAgent() for i in range(10)]
            

    def iter(n): # one model iteration, repeated n times
        my_agents = World().get_agents_of_type("MyAgent")

        vals = []
        for i in range(n):
            # write agents' parameters here (just an example)
            [agent.modify_a() for agent in my_agents]
            a_param_vals = [agent.a_param for agent in my_agents]
            vals.append(np.mean(a_param_vals))

        return pd.DataFrame({"Value":vals}) # has to return a dataframe

    # create model runner
    print("settings path", settings_path)
    mr = ModelRunner(settings_path,results_path,builder,iter)
    mr.run(10,20)

    # read back the output files
    with open(os.path.join(results_path, "output.txt"), "r") as file:
        print("1. output.txt:\n ", file.readlines()[:3], "\n")

    with open(os.path.join(results_path, "progress.txt"), "r") as file:
        print("2. progress.txt:\n ", file.readlines()[:3], "\n")

    with open(os.path.join(results_path, "output.0"), "r") as file:
        print("3. values:\n ", file.readlines()[:3], "\n")



class MonteCarloExample(Example):
    def __init__(self, base_path = None, settings_path = None, results_path = None):

        # assign default paths 
        if base_path is None:
            base_path = "sfctools/examples/monte_carlo/"

        super().__init__(lambda : run(base_path))


if __name__ == "__main__":

    base_path = os.getcwd()
    mc_ex = MonteCarloExample(base_path) 
    mc_ex.run()


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

from ..core.settings import Settings
from ..core.world import World
from ..core.flow_matrix import FlowMatrix
from pathlib import Path

import sys
import os
from datetime import datetime
import warnings
import pandas as pd


class ModelRunner:
    """
    A model runner is a generic 'frame' for running a model file. It requires a specific folder structure (see constructor).
    The usage of this is optional and voluntary. However, this could be helpful when running Monte-Carlo batches.
    """

    def __init__(self, settings_path, results_path, builder, iter):
        """

        :param settings_path: str or path to .yml file
        :param results_path: str or path, dir where results are written
        :param builder: callable, method to call when building agents
        :param iter: callable, method to call when iterating one simulation, should have an argument N (for number of iterations) and must return a dataframe (or a comparable data structure which supports a to_csv method)

        This runs a simulation and create logger files and data files

            | ./root
            | |_____output.txt  <<
            | |_____errors.txt << will contain all error messages
            | |_____progress.txt << will log the progress of the simulation run
            | |
            | |
            | |_____output.0
            | |_____output.1
            | |_____output.2
            | |_____output.3 << will contain the data of the actual runs


        """

        """
        below, we write some header information into some logger files
        The .pogress file
        The .errors file
        """

        Path(results_path).mkdir(parents=True, exist_ok=True)

        Settings().read_from_yaml(settings_path)  # read settings from yaml file

        with open(results_path + "output.txt", "w") as file:  # write the meta-info to results directory
            info = Settings().config_data["metainfo"]
            for k, v in info.items():
                file.write(str(k) + ":" + str(v) + "\n")

        with open(results_path + "progress.txt", "w") as file:  # write timestamp to progress logger file
            file.write(str(datetime.now()) + "\n")

        with open(results_path + "errors.txt", "w") as file:  # write timestamp to error file
            file.write(str(datetime.now()) + "\n")

        self.build_agents = builder
        self.iter = iter
        self.settings_path = settings_path
        self.results_path = results_path

    def run(self, M: int, N: int):
        """
        runs a monte-carlo simulation wih M monte carlo runs and N simulation periods per run.
        Outputs are written to logger files.

        :param M: number of monte-carlo runs
        :param N: number of time periods per monte-carlo run
        """

        sys.tracebacklimit = 1

        error_count = 0

        m = 0
        while m < M:

            with open(self.results_path + "progress.txt", "a") as file:
                file.write("[RUN %i]   " % m)

            try:

                World().reset()
                FlowMatrix().reset()

                self.build_agents()  # build the agents

                data = self.iter(N)  # iterate one simulation run

                # sys.stdout.flush()
                # sys.stderr.close()

                data.to_csv(self.results_path + "output.%i" % m,sep="\t",index=False)

                sys.stdout.write("[DATA]" + self.results_path + ".%i\n" % m)

                with open(self.results_path + "output.txt", "a") as file:  # write-append
                    # logging here
                    file.write(self.results_path + "output.%i\n" % m)

                with open(self.results_path + "progress.txt", "a") as file:
                    file.write("...SUCCESS!\n")

                sys.stdout.write("[UPDATE]%.6f\n" % (100. * m / M))

                m += 1

            except Exception as e:

                with open(self.results_path + "progress.txt", "a") as file:
                    file.write("...FAILED\n")

                with open(self.results_path + "errors.txt", "a") as file:

                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                    e_str = ":".join([str(exc_type), str(fname), str(exc_tb.tb_lineno)]) +" "+ str(e)

                    file.write("[Error in run %i] " % m + e_str + "\n")

                    warnings.warn("[Error in run %i] " % m + e_str + "\n")

                error_count += 1
                if error_count == 10:
                    error = RuntimeError("ᵒᴼᵒ ✈__✈ █ █ ▄  Got a lot of errors. aborting. something must be wrong, see error logger file.")
                    raise error



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod


from sfctools import MarketMatching, Agent
import numpy as np
np.random.seed(1234)
from sfctools.examples.example_wrapper import Example


def run(show_plot=True):
    # define a rudimentary trader class

    class Trader(Agent):
        # a rudimentary trader
        def __init__(self):
            super().__init__()

    supply_traders = [Trader() for i in range(50)]
    demand_traders = [Trader() for i in range(22)] # generate some agents

    # define a market and a matching rule

    class Market(MarketMatching):
        """
        My market for matching supply and demand
        """
        def __init__(self):
            super().__init__()

        def rematch(self):
            # match the suppliers and demanders at random

            for i in self.demand_list:
                for j in self.supply_list:
                    u = np.random.rand()
                    if u > 0.3:
                        self.link_agents(j,i,u)


    # generate a market and add the traders
    my_market = Market()
    [my_market.add_demander(agent) for agent in demand_traders]
    [my_market.add_supplier(agent) for agent in supply_traders]

    # re-match the market traders
    my_market.rematch()

    # plot the resulting network
    if show_plot:
        my_market.plot()

    # print suppliers of demander 0
    print(my_market.get_suppliers_of(demand_traders[0]))


class MarketExample(Example):
    def __init__(self,show_plot=False):
        super().__init__(lambda: run(show_plot=show_plot))


if __name__ == "__main__":
    my_instance = MarketExample(show_plot=True)
    my_instance.run()



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example


from .agents.household import Household
from .agents.production import Production
from .agents.government import Government

from sfctools import Clock, FlowMatrix, Settings, World
from sfctools.misc.mpl_plotting import matplotlib_lineplot, matplotlib_barplot
from sfctools import IndicatorReport, ReportingSheet, BalanceSheet

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np


def run(show_plot=True):

    FlowMatrix().reset()
    World().reset()


    # read settings
    Settings().read_from_yaml("sfctools/examples/exampleabm/settings.yml")

    # generate agents
    my_household  = Household()
    my_production = Production()
    my_government = Government()

    # link the agents
    my_household.link()
    my_production.link()
    my_government.link()

    # simulation duration
    T = int(Settings().get_hyperparameter("T")) # simulation length
    T_burn_in = int(Settings().get_hyperparameter("T_burn")) # simulation length

    # prepare output data
    rs = ReportingSheet([
        IndicatorReport("Time","C"),
        IndicatorReport("Time","G"),
        IndicatorReport("Time","Y"),
        IndicatorReport("Time","T"),
        IndicatorReport("Time","DI")])

    household_balance_data = []  # logs the household balance sheets
    governmnt_balance_data = []  # logs the government balance sheets

    # simulation loop
    for i in range(T):

        # reset flow matrix
        FlowMatrix().reset()

        # update all agents individually
        my_household.update()
        my_government.update()
        my_production.update()

        Clock().tick() # increase clock by one tick

        # output data
        if i > T_burn_in: # after burn-in

            IndicatorReport.getitem("C").add_data(my_household.C)
            IndicatorReport.getitem("G").add_data(my_government.G)
            IndicatorReport.getitem("Y").add_data(my_production.Y)
            IndicatorReport.getitem("T").add_data(my_production.T)
            IndicatorReport.getitem("DI").add_data(my_household.DI)

            # plot agent's balance sheets
            household_balance_data.append(my_household.balance_sheet.raw_data)
            governmnt_balance_data.append(my_government.balance_sheet.raw_data)

        # reset agents
        my_household.reset()
        my_production.reset()
        my_government.reset()

    # print flow matrix
    print(FlowMatrix().to_string())

    if show_plot:
        fig = FlowMatrix().plot_sankey(show_values=True,show_plot=False)
        plt.savefig("sfctools/examples/exampleabm/figures/sankey.png")
        plt.close()

        fig = FlowMatrix().plot_colored(show_plot=False)
        plt.savefig("sfctools/examples/exampleabm/figures/matrix.png")
        plt.close()

        # plot output indicators
        fig = rs.plot(show_plot=False)
        plt.savefig("sfctools/examples/exampleabm/figures/outputs.png")
        plt.close()

        # plot the stocks of government and household
        # BalanceSheet.plot_list(household_balance_data, dt=5, xlabel="Time", title="Household",  show_liabilities=False)
        BalanceSheet.plot_list(governmnt_balance_data, dt=5, xlabel="Time", title="Example", show_liabilities=True)


class ABMExample(Example):
    def __init__(self):
        super().__init__(lambda: run(show_plot=False))


if __name__ == "__main__":
    my_instance = ABMExample()
    my_instance.run()
